# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause
from collections import defaultdict
from itertools import accumulate
from typing import Any, Generator

from bitstring import ConstBitStream, Error
from bitstring.bits import BitsType
from sortedcontainers import SortedList

from ..keyframeinfo import KeyframeInfo

box_atoms = {"moov", "trak", "mdia", "minf", "stbl", "edts"}  # Non-exhaustive


def parse_table(cbs: ConstBitStream, table_size: int, struct: dict[str, str]) -> dict[str, Any]:
    return [
        dict(zip(struct.keys(), cbs.readlist(", ".join(struct.values()))))
        for _ in range(table_size)
    ]


class Atom:
    skip_version_and_flags: bool = False

    @staticmethod
    def make_atom(cbs: ConstBitStream) -> "Atom":
        size: int = cbs.read("uint:32")
        name: str = cbs.read("bytes:4").decode("ascii")
        box: bool = name in box_atoms

        if size == 0:
            raise RuntimeError(
                "MPEG parser detected a zero byte atom, this likely indicates a corrupt video."
            )

        subclass_list = [c for c in Atom.__subclasses__() if c.__name__ == name.upper()]
        atom_class: type = Atom
        if len(subclass_list) > 0:
            atom_class: type = subclass_list[0]
            cbs.bytepos += 4  # Skip version and flags TODO not every atom needs this

        atom = atom_class(size, name, box)
        atom._parse(cbs)

        return atom

    def __init__(self, size: int, name: str, box: bool) -> None:
        self.size: int = size
        self.name: str = name
        self.box: bool = box

    def _parse(self, cbs: ConstBitStream) -> None:
        if not self.box:
            cbs.bytepos += self.size - 8

    def __str__(self) -> str:
        return f"{self.name=}, {self.size=}, {self.box=}"


class TKHD(Atom):
    """
    Parses the track header atom, see https://developer.apple.com/documentation/quicktime-file-format/track_header_atom
    """

    def _parse(self, cbs: ConstBitStream) -> None:
        cbs.bytepos += 8  # skip creation time and modification time
        self.track_id: int = cbs.read("uint:32")
        cbs.bytepos += 68  # Skip rest of structure


class HDLR(Atom):
    """
    Parses the media handler atom, see https://developer.apple.com/documentation/quicktime-file-format/handler_reference_atom

    NOTE: currently unused but could speed up parsing by skipping audio tracks
    """

    def _parse(self, cbs: ConstBitStream) -> None:
        self.component_type = cbs.read("bytes:4").decode("ascii")
        self.component_subtype = cbs.read("bytes:4").decode("ascii")

        # Skip rest of structure, the last field is variable so we need to use the total size
        # 24 bytes already read (size (4), type (4), version (1), flags (3), component type (4), component subtype (4))
        cbs.bytepos += self.size - 20


class STSS(Atom):
    """
    Parses the sync sample atom https://developer.apple.com/documentation/quicktime-file-format/sample_table_atom/sync_sample_atom
    """

    def _parse(self, cbs: ConstBitStream) -> None:
        self.number_of_entries: int = cbs.read("uint:32")
        self.sync_sample_table: dict[str, Any] = parse_table(
            cbs, self.number_of_entries, {"number": "uint:32"}
        )


class STTS(Atom):
    """
    Parses the time to sample atom https://developer.apple.com/documentation/quicktime-file-format/time-to-sample_atom
    """

    def _parse(self, cbs: ConstBitStream) -> None:
        self.number_of_entries: int = cbs.read("uint:32")
        self.time_to_sample_table: dict[str, Any] = parse_table(
            cbs,
            self.number_of_entries,
            {"sample_count": "uint:32", "sample_duration": "uint:32"},
        )


class CTTS(Atom):
    """
    Parses the composition offset atom https://developer.apple.com/documentation/quicktime-file-format/composition_offset_atom
    """

    def _parse(self, cbs: ConstBitStream) -> None:
        self.number_of_entries: int = cbs.read("uint:32")
        self.composition_offset_table: dict[str, Any] = parse_table(
            cbs,
            self.number_of_entries,
            {
                "sample_count": "uint:32",
                "composition_offset": "int:32",
                "media_rate": "",
            },
        )


class ELST(Atom):
    """
    Parses the edit list atom https://developer.apple.com/documentation/quicktime-file-format/edit_list_atom
    """

    def _parse(self, cbs: ConstBitStream) -> None:
        self.number_of_entries: int = cbs.read("uint:32")
        self.edit_list_table: dict[str, Any] = parse_table(
            cbs,
            self.number_of_entries,
            {
                "track_duration": "uint:32",
                "media_time": "int:32",
                "media_rate": "int:32",
            },
        )


class MDAT(Atom):
    """
    Parses the media data atom https: https://developer.apple.com/documentation/quicktime-file-format/movie_data_atom

    This is only here to handle the unusual size handling of mdat, if the normal size field is set to 1
    then the actual size is stored as a 64 bit integer
    """

    def _parse(self, cbs: ConstBitStream) -> None:
        if self.size == 1:
            cbs.bytepos -= 4  # No version or flags for mdat
            self.size = cbs.read("uint:64")
            seekto = self.size - 16
        else:
            seekto = self.size - 12

        if cbs.bytepos + seekto >= (cbs.len / 8):
            raise StopIteration()

        cbs.bytepos += seekto


def parse_atoms(file: BitsType) -> Generator[Atom, None, None]:
    try:
        cbs = ConstBitStream(file)
        while cbs.pos < len(cbs):
            try:
                yield Atom.make_atom(cbs)
            except StopIteration:
                return
    except Error as e:
        raise ValueError(f"MPEG parsing failed with error {e}")


def parse_mpeg(file: BitsType) -> dict[int, SortedList]:
    sync_samples = {}
    decode_timestamps = {}
    presentation_time_offsets = defaultdict(lambda: defaultdict(lambda: 0))
    start_offsets = defaultdict(lambda: 0)
    current_track = -1
    for a in parse_atoms(file):
        if a.name == "tkhd":
            a: TKHD
            current_track = a.track_id
        elif a.name == "stts":
            a: STTS
            decode_timestamps[current_track] = list(
                accumulate(
                    sum(
                        [
                            [entry["sample_duration"]] * entry["sample_count"]
                            for entry in a.time_to_sample_table
                        ],
                        [0],
                    )
                )
            )
        elif a.name == "ctts":
            a: CTTS
            presentation_time_offsets[current_track] = sum(
                [
                    [entry["composition_offset"]] * entry["sample_count"]
                    for entry in a.composition_offset_table
                ],
                [],
            )
        elif a.name == "stss":
            a: STSS
            sync_samples[current_track] = [ss["number"] - 1 for ss in a.sync_sample_table]
        elif a.name == "elst":
            # NOTE the "media_time" here is a "delay" between decoding and presenting the first sample.
            # We follow the ffmpeg convention that the first frame displays at time 0 which means we should
            # *subtract* this offset from the decoding time values rather than adding it to presentation time values
            # TODO there can be more than one of these, figure out how to handle it
            a: ELST
            start_offsets[current_track] = -a.edit_list_table[0]["media_time"]
    keyframes = defaultdict(lambda: SortedList())
    try:
        for track_id in sync_samples.keys():
            for keyframe_number in sync_samples[track_id]:
                pts = (
                    decode_timestamps[track_id][keyframe_number]
                    + start_offsets[track_id]
                    + presentation_time_offsets[track_id][keyframe_number]
                )
                keyframes[track_id].add(KeyframeInfo(keyframe_number, pts))
    except (KeyError, IndexError) as e:
        raise ValueError(f"MPEG parsing failed with error {e}")

    return keyframes



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
from typing import Any, Callable, Dict, Generator, Generic, Iterator, List, Tuple, TypeVar, Union

import torch

from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.base import BaseWrapperDataset
from megatron.energon.wrappers.batch_dataset import BatchDataset
from megatron.energon.wrappers.blend_dataset import BlendDataset

T_batch_in = TypeVar("T_batch_in")
T_batch = TypeVar("T_batch")


def generic_concat(batch: List[Any]) -> Any:
    """Based on the types/shapes of the batch: Will either pad and stack, or return as list.
    Recurses structures (dict, dataclass, namedtuple) and applies the same logic to each field."""
    if isinstance(batch[0], torch.Tensor):
        return concat_pad(batch)
    elif isinstance(batch[0], dict):
        return {key: generic_concat([sample[key] for sample in batch]) for key in batch[0].keys()}
    elif dataclasses.is_dataclass(batch[0]):
        return type(batch[0])(
            **{
                field.name: generic_concat([getattr(sample, field.name) for sample in batch])
                for field in dataclasses.fields(batch[0])
            }
        )
    elif isinstance(batch[0], tuple) and hasattr(batch[0], "_fields"):
        # NamedTuple
        return type(batch[0])(
            **{
                field: generic_concat([getattr(sample, field) for sample in batch])
                for field in batch[0]._fields
            }
        )
    else:
        return batch


def concat_pad(batch: List[Any]) -> Any:
    """Concat a batch of arbitrary-sized tensors padded with 0s."""
    total_bs = sum(b.shape[0] for b in batch)
    max_size = [max(b.shape[dim] for b in batch) for dim in range(1, batch[0].ndim)]
    concat_tensor = batch[0].new_zeros((total_bs, *max_size))
    b_idx = 0
    for b in batch:
        concat_tensor[(slice(b_idx, b_idx + b.shape[0]), *(slice(0, s) for s in b.shape[1:]))] = b
        b_idx += b.shape[0]
    # Pad all tensors to max_size
    return concat_tensor


def homogeneous_concat_mix(samples: List[T_batch_in]) -> T_batch:
    """
    Mixes a list of batches into a single batch. The default implementation is to concat the
    batches if they are all of the same type, otherwise return a list of batches.

    Args:
        samples: THe samples to mix.

    Returns:
        The mixed batch.
    """
    first_type = type(samples[0])
    assert all(first_type is type(sample) for sample in samples)
    # All the same type -> concat batches
    return generic_concat(samples)


class MixBatchDataset(BaseWrapperDataset[T_batch_in, T_batch], Generic[T_batch_in, T_batch]):
    """
    This dataset wrapper blends multiple iterable datasets together give a weight.
    The datasets may be infinite. This dataset is always infinite.
    Effectively combines :class:`megatron.energon.BlendDataset` and :class:`megatron.energon.BatchDataset`.
    """

    def __init__(
        self,
        *dataset_weights: Tuple[SavableDataset[T_batch_in], float],
        batch_size: int,
        batch_mix_fn: Callable[
            [List[T_batch_in]], Union[T_batch, Generator[T_batch, None, None]]
        ] = lambda x: x,
        worker_config: WorkerConfig,
    ):
        """Construct a BlendDataset.

        Args:
            dataset_weights: Each argument should be a tuple of (dataset, weight) with a weight
                between 0 and 1. The output samples are sampled from the input datasets with the
                given probabilities. The datasets should have a batch size of 1, otherwise the
                whole batches will be sampled.
            batch_size: The batch size to output.
            batch_mix_fn: A function that takes a list of samples from the input datasets and
                returns a batch sample. The default implementation returns a list of batches.
                For homogeneous datasets, it is recommended to use the
                :func:`megatron.energon.homogeneous_concat_mix` which concatenates the batches. May raise
                :exc:`megatron.energon.SkipSample` to skip a sample. May also return a generator, which
                will be iterated over to produce batches.
            worker_config: Configuration for the workers.
        """
        super().__init__(
            BatchDataset(
                BlendDataset(*dataset_weights, worker_config=worker_config),
                batch_size=batch_size,
                batcher=batch_mix_fn,
                worker_config=worker_config,
            ),
            worker_config=worker_config,
        )

    def reset_state_own(self) -> None:
        return

    def __len__(self) -> int:
        return len(self.dataset)

    def __iter__(self) -> Iterator[T_batch]:
        yield from self.dataset

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
        }

    def __str__(self):
        return f"MixBatchDataset(dataset={self.dataset})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause
from collections import defaultdict

from bitstring.bits import BitsType
from ebmlite import MasterElement, loadSchema
from sortedcontainers import SortedList

from ..keyframeinfo import KeyframeInfo


class CueTrackPositions:
    track: int

    def __init__(self, el: MasterElement) -> None:
        for c in el:
            if c.name == "CueTrack":
                self.track = c.value


class CuePoint:
    time: int
    track_positions: CueTrackPositions

    def __init__(self, el: MasterElement) -> None:
        for c in el:
            if c.name == "CueTime":
                self.time = c.value
            if c.name == "CueTrackPositions":
                self.track_positions = CueTrackPositions(c)


def parse_matroska(file: BitsType) -> SortedList:
    try:
        schema = loadSchema("matroska.xml")
        doc = schema.load(file, headers=True)
    except (KeyError, IOError, TypeError) as e:
        raise ValueError(f"Matroska parsing failed with error {e}")

    # Get cue times
    stack = [c for c in doc if c.name == "Segment"]
    cues = defaultdict(SortedList)
    while len(stack) > 0:
        el = stack.pop()
        if el.name == "CuePoint":
            cue = CuePoint(el)
            cues[cue.track_positions.track].add(KeyframeInfo(cue.time, cue.time))
        elif isinstance(el, MasterElement):
            stack.extend([c for c in el if c.name in ["Cues", "CuePoint"]])

    return cues



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import (
    Any,
    Dict,
    Generator,
    Generic,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from megatron.energon.flavors.base_dataset import FlexState, SavableDataset
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.base import BaseWrapperDataset, get_sample_restore_key

T_sample = TypeVar("T_sample")


class SavableSampleBuffer(BaseWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """A buffer of samples, savable."""

    _buffer: List[T_sample]
    _restore_keys: List[Tuple[Union[str, int, tuple], ...]]

    _savable_fields = ("_restore_keys",)
    _restore_pending: bool = False

    def __init__(self, dataset: SavableDataset[T_sample], *, worker_config: WorkerConfig):
        super().__init__(dataset, worker_config=worker_config)
        self.reset_state_own()

    def reset_state_own(self) -> None:
        self._buffer = []
        self._restore_keys = []

    def worker_start(self) -> None:
        if self._restore_pending:
            assert len(self._buffer) == 0
            self._restore_pending = False
            for restore_key in self._restore_keys:
                self._buffer.append(self.restore_sample(restore_key))
        assert len(self._buffer) == len(self._restore_keys)

    def append(self, sample: T_sample) -> T_sample:
        self._buffer.append(sample)
        self._restore_keys.append(get_sample_restore_key(sample))
        return sample

    def extend(self, samples: List[T_sample], restore_keys: Optional[Sequence[Any]] = None) -> None:
        self._buffer.extend(samples)
        if restore_keys is None:
            self._restore_keys.extend(get_sample_restore_key(sample) for sample in samples)
        else:
            self._restore_keys.extend(restore_keys)

    def append_iter(self) -> Generator[T_sample, None, None]:
        for sample in self.dataset:
            yield self.append(sample)

    def pop(self, index: int) -> T_sample:
        self._restore_keys.pop(index)
        return self._buffer.pop(index)

    def flush(self) -> Tuple[List[T_sample], Tuple[Any, ...]]:
        buffer = list(self._buffer)
        restore_key = tuple(self._restore_keys)
        self._buffer.clear()
        self._restore_keys.clear()
        return buffer, restore_key

    def __iter__(self) -> Iterator[T_sample]:
        return iter(self._buffer)

    def __getitem__(self, index: Union[int, slice]) -> Union[T_sample, List[T_sample]]:
        return self._buffer[index]

    def __setitem__(self, index: Union[int, slice], value: T_sample) -> None:
        self._buffer[index] = value
        if isinstance(index, slice):
            self._restore_keys[index] = (get_sample_restore_key(v) for v in value)
        else:
            self._restore_keys[index] = get_sample_restore_key(value)

    def __delitem__(self, index: Union[int, slice]) -> None:
        del self._buffer[index]
        del self._restore_keys[index]

    def __len__(self) -> int:
        return len(self._restore_keys)

    def save_state(self) -> FlexState:
        # Don't call super().save_state() because we don't want to save the wrapped datasets
        # Just save the own state
        return SavableDataset.save_state(self)

    def restore_state(self, state: FlexState) -> None:
        # Don't call super().restore_state() because we don't want to restore the wrapped datasets
        # Just restore the own state
        SavableDataset.restore_state(self, state)

        self._restore_pending = True

    def restore_key(self) -> Tuple[Union[str, int], ...]:
        return tuple(self._restore_keys)

    def restore_samples(
        self, index: Tuple[Union[str, int, tuple], ...]
    ) -> Tuple[Tuple[Union[str, int, tuple], ...], List[T_sample]]:
        buffer = []
        restore_keys = []
        for sub_index in index:
            sample = self.restore_sample(sub_index)
            restore_keys.append(get_sample_restore_key(sample))
            buffer.append(sample)
        return tuple(restore_keys), buffer

    def clear(self) -> None:
        self._buffer.clear()
        self._restore_keys.clear()

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "worker_config": self.worker_config.config(),
        }

    def debug_print(self, indent: str = ""):
        print(
            f"{indent}SavableSampleBuffer(size={len(self._restore_keys)}, res_pend={self._restore_pending}):\n",
            end="",
        )
        for i, (sample, restore_key) in enumerate(zip(self._buffer, self._restore_keys)):
            print(f"{indent}Sample {i} [{restore_key!r}]: {sample.__key__}\n", end="")

    def __str__(self):
        return f"SavableSampleBuffer(size={len(self._buffer)})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Callable, ClassVar, Dict, Optional, Union

from megatron.energon.epathlib import EPath


class ModuleLoader:
    """Class providing access to python files to be loaded as modules."""

    #: Loaded modules. Global singleton, no need to load a module multiple times.
    _modules: ClassVar[Dict[EPath, Dict[str, Any]]] = {}

    def _load_module(self, path: Union[str, EPath]):
        """Internally loads a module"""
        path = EPath(path)
        with path.open("r") as rf:
            src_code = rf.read()
        code = compile(src_code, path.url, "exec")
        mod = {}
        eval(code, mod)
        self._modules[path] = mod

    def __getitem__(self, path: Union[str, EPath]) -> Any:
        """
        Gets a module by path.

        Args:
            path: The path to load

        Returns:
            Loaded module dict.
        """
        path = EPath(path)
        if path not in self._modules:
            self._load_module(path)
        return self._modules[path]

    def get_function(
        self,
        module_fn: str,
        default_fn_name: Optional[str] = None,
        relative_path: Optional[EPath] = None,
    ) -> Callable:
        """
        Loads a function in the form `"module/submodule.py:function_name"`.

        Args:
            module_fn: The function string to load. Must contain the path of the python module
                and a function name separated by a colon (might be omitted if `default_fn_name` is
                specified).
            default_fn_name: Default name of the function if not given in `module_fn` string.
            relative_path: The relative parent path to the module. If not specified, the current
                working directory / absolute path is used.

        Returns:
            The function from the module
        """
        if ":" in module_fn:
            module, fn_name = module_fn.rsplit(":", 1)
        else:
            if default_fn_name is None:
                raise ValueError("Function name must be specified")
            fn_name = default_fn_name
            module = module_fn
        if relative_path is not None:
            module = str(relative_path / module)
        fn = self[module][fn_name]
        assert callable(fn)
        return fn



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import math
import numbers
from abc import ABC, abstractmethod
from typing import Any, ClassVar, Generic, List, Optional, Tuple, Type, Union

import numpy as np
import torch
from torchvision import transforms as T

from megatron.energon.transforms.common import TTransform


class TransformMapper(ABC, Generic[TTransform]):
    source_type: ClassVar[Type[torch.nn.Module]]

    transform: TTransform

    def __init__(self, transform: TTransform):
        self.transform = transform

    def __call__(self, sample):
        return self.transform(sample)

    @staticmethod
    def translate(x: float, y: float) -> np.ndarray:
        m = np.eye(3, dtype=np.float64)
        m[0, 2] = x
        m[1, 2] = y
        return m

    @staticmethod
    def rotate(angle: float) -> np.ndarray:
        """Counter-clockwise rotation. Note that the Y-axis is point down."""
        m = np.eye(3, dtype=np.float64)
        m[:2, :2] = np.array([[np.cos(angle), np.sin(angle)], [-np.sin(angle), np.cos(angle)]])
        return m

    @staticmethod
    def scale(x: float, y: float) -> np.ndarray:
        m = np.eye(3, dtype=np.float64)
        m[0, 0] = x
        m[1, 1] = y
        return m

    @staticmethod
    def shear(x: float, y: float) -> np.ndarray:
        m = np.eye(3, dtype=np.float64)
        m[0, 1] = x
        m[1, 0] = y
        return m

    @staticmethod
    def hflip() -> np.ndarray:
        m = np.eye(3, dtype=np.float64)
        m[0, 0] = -1
        return m

    @staticmethod
    def vflip() -> np.ndarray:
        m = np.eye(3, dtype=np.float64)
        m[1, 1] = -1
        return m

    @abstractmethod
    def apply_transform(
        self, matrix: np.ndarray, dst_size: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, Any]: ...

    def fill(
        self,
    ) -> Optional[Union[int, float, Tuple[Union[int, float], ...], List[Union[int, float]]]]:
        return None

    def interpolation(self) -> Optional[T.InterpolationMode]:
        return None


class ResizeMapper(TransformMapper[T.Resize]):
    source_type = T.Resize

    def __init__(self, transform: T.Resize):
        super().__init__(transform)

    def _compute_resized_output_size(
        self, image_size: Tuple[int, int], size: List[int], max_size: Optional[int] = None
    ) -> List[int]:
        if len(size) == 1:  # specified size only for the smallest edge
            h, w = image_size
            short, long = (w, h) if w <= h else (h, w)
            requested_new_short = size[0]

            new_short, new_long = requested_new_short, int(requested_new_short * long / short)

            if max_size is not None:
                if max_size <= requested_new_short:
                    raise ValueError(
                        f"max_size = {max_size} must be strictly greater than the requested "
                        f"size for the smaller edge size = {size}"
                    )
                if new_long > max_size:
                    new_short, new_long = int(max_size * new_short / new_long), max_size

            new_w, new_h = (new_short, new_long) if w <= h else (new_long, new_short)
        else:  # specified both h and w
            new_w, new_h = size[1], size[0]
        return [new_h, new_w]

    def apply_transform(
        self, matrix: np.ndarray, dst_size: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, Tuple[Any, ...]]:
        size = self.transform.size

        if isinstance(size, int):
            size = [size]

        h, w = self._compute_resized_output_size(dst_size, size, self.transform.max_size)

        matrix = self.scale(w / dst_size[1], h / dst_size[0]) @ matrix
        # matrix = self.scale((w - 1) / (dst_size[1] - 1), (h - 1) / (dst_size[0] - 1)) @ matrix
        # matrix = self.translate(0.25, 0.25) @ matrix
        # matrix = self.translate(0.1, 0) @ matrix
        dst_size = np.array((h, w), dtype=dst_size.dtype)
        # print(f"Resize s={size}")
        return matrix, dst_size, (self.source_type.__name__, size)

    def interpolation(self) -> Optional[T.InterpolationMode]:
        return self.transform.interpolation


class RandomResizedCropMapper(TransformMapper[T.RandomResizedCrop]):
    source_type = T.RandomResizedCrop

    def __init__(self, transform: T.RandomResizedCrop):
        super().__init__(transform)

    def get_params(self, size: Tuple[int, int]) -> Tuple[int, int, int, int]:
        """
        Gets the parameters for a random resized crop.
        This function is derived from T.RandomResizedCrop.get_params, but without requiring the
        input image (to determine the input size).

        Returns:
            Tuple of (top, left, height, width).
        """
        height, width = size
        area = height * width

        log_ratio = torch.log(torch.tensor(self.transform.ratio))
        for _ in range(10):
            target_area = (
                area
                * torch.empty(1).uniform_(self.transform.scale[0], self.transform.scale[1]).item()
            )
            aspect_ratio = torch.exp(torch.empty(1).uniform_(log_ratio[0], log_ratio[1])).item()

            w = int(round(math.sqrt(target_area * aspect_ratio)))
            h = int(round(math.sqrt(target_area / aspect_ratio)))

            if 0 < w <= width and 0 < h <= height:
                i = torch.randint(0, height - h + 1, size=(1,)).item()
                j = torch.randint(0, width - w + 1, size=(1,)).item()
                return i, j, h, w

        # Fallback to central crop
        in_ratio = float(width) / float(height)
        if in_ratio < min(self.transform.ratio):
            w = width
            h = int(round(w / min(self.transform.ratio)))
        elif in_ratio > max(self.transform.ratio):
            h = height
            w = int(round(h * max(self.transform.ratio)))
        else:  # whole image
            w = width
            h = height
        i = (height - h) // 2
        j = (width - w) // 2
        return i, j, h, w

    def apply_transform(
        self, matrix: np.ndarray, dst_size: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, Tuple[Any, ...]]:
        top, left, height, width = self.get_params(dst_size)
        # print(
        #     "RandomResizedCrop", top, left, dst_size[0] - height - top, dst_size[1] - width - left
        # )
        # Crop to left, top, height, width
        matrix = self.translate(-left, -top) @ matrix
        dst_size = np.array([height, width], dtype=dst_size.dtype)
        # Resize to target size
        matrix = (
            self.scale(self.transform.size[1] / dst_size[1], self.transform.size[0] / dst_size[0])
            @ matrix
        )
        dst_size = np.array(self.transform.size, dtype=dst_size.dtype)
        return matrix, dst_size, (self.source_type.__name__, (top, left, height, width))

    def interpolation(self) -> Optional[T.InterpolationMode]:
        return self.transform.interpolation


class RandomHorizontalFlipMapper(TransformMapper[T.RandomHorizontalFlip]):
    source_type = T.RandomHorizontalFlip

    def __init__(self, transform: T.RandomHorizontalFlip):
        super().__init__(transform)

    def apply_transform(
        self, matrix: np.ndarray, dst_size: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, Any]:
        do_flip = torch.rand(1) < self.transform.p
        if do_flip:
            matrix = self.hflip() @ matrix
            matrix = self.translate(dst_size[1], 0) @ matrix
            # print(f"RandomHorizontalFlip")
        return matrix, dst_size, (self.source_type.__name__, do_flip)


class RandomVerticalFlipMapper(TransformMapper[T.RandomVerticalFlip]):
    source_type = T.RandomVerticalFlip

    def __init__(self, transform: T.RandomVerticalFlip):
        super().__init__(transform)

    def apply_transform(
        self, matrix: np.ndarray, dst_size: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, Any]:
        do_flip = torch.rand(1) < self.transform.p
        if do_flip:
            matrix = self.vflip() @ matrix
            matrix = self.translate(0, dst_size[0]) @ matrix
            # print(f"RandomVerticalFlip")
        return matrix, dst_size, (self.source_type.__name__, do_flip)


class RandomRotationMapper(TransformMapper[T.RandomRotation]):
    source_type = T.RandomRotation

    def __init__(self, transform: T.RandomRotation):
        super().__init__(transform)

    def apply_transform(
        self, matrix: np.ndarray, dst_size: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, Any]:
        assert self.transform.center is None, "Only centered rotation is supported"
        degrees = self.transform.get_params(self.transform.degrees)
        rads = degrees * np.pi / 180
        # print(f"Rotate deg={degrees}")
        orig_size = dst_size
        if self.transform.expand:
            # Compute size of rotated rectangle
            w = np.abs(np.sin(rads)) * dst_size[0] + np.abs(np.cos(rads)) * dst_size[1]
            h = np.abs(np.sin(rads)) * dst_size[1] + np.abs(np.cos(rads)) * dst_size[0]

            # Round in the same way as PIL does
            rounded_w = np.ceil(orig_size[1] / 2 + w / 2) - np.floor(orig_size[1] / 2 - w / 2)
            rounded_h = np.ceil(orig_size[0] / 2 + h / 2) - np.floor(orig_size[0] / 2 - h / 2)

            # New size is h, w
            dst_size = np.array([int(rounded_h), int(rounded_w)], dtype=dst_size.dtype)
        matrix = (
            self.translate(dst_size[1] / 2, dst_size[0] / 2)
            @ self.rotate(rads)
            @ self.translate(-orig_size[1] / 2, -orig_size[0] / 2)
            @ matrix
        )
        return matrix, dst_size, (self.source_type.__name__, degrees)

    def fill(
        self,
    ) -> Optional[Union[int, float, Tuple[Union[int, float], ...], List[Union[int, float]]]]:
        return self.transform.fill

    def interpolation(self) -> Optional[T.InterpolationMode]:
        return self.transform.interpolation


class RandomCropMapper(TransformMapper[T.RandomCrop]):
    source_type = T.RandomCrop

    def __init__(self, transform: T.RandomCrop):
        super().__init__(transform)

    def apply_transform(
        self, matrix: np.ndarray, dst_size: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, Any]:
        th, tw = self.transform.size  # Target height and width

        # pad the width if needed
        if self.transform.pad_if_needed and dst_size[1] < tw:
            padding = tw - dst_size[1]  # Pad this much on both left and right
            matrix = self.translate(padding, 0) @ matrix
            dst_size[1] += 2 * padding

        # pad the height if needed
        if self.transform.pad_if_needed and dst_size[0] < th:
            padding = th - dst_size[0]  # Pad this much on both top and bottom
            matrix = self.translate(0, padding) @ matrix
            dst_size[0] += 2 * padding

        h, w = dst_size

        if h < th or w < tw:
            raise ValueError(
                f"Required crop size {(th, tw)} is larger than input image size {(h, w)}"
            )

        if w == tw and h == th:
            # No need to crop if we're at the target size already
            i = 0
            j = 0
        else:
            i = torch.randint(0, h - th + 1, size=(1,)).item()  # Offset y
            j = torch.randint(0, w - tw + 1, size=(1,)).item()  # Offset x
            matrix = self.translate(-j, -i) @ matrix

        if self.transform.pad_if_needed:
            dst_size = np.array((th, tw), dtype=dst_size.dtype)
        else:
            dst_size = np.array((min(th, dst_size[0]), min(tw, dst_size[1])), dtype=dst_size.dtype)
        # print(f"RandomCrop t=[{dx}, {dy}], s={dst_size}")
        return matrix, dst_size, (self.source_type.__name__, (j, i, th, tw))

    def fill(
        self,
    ) -> Optional[Union[int, float, Tuple[Union[int, float], ...], List[Union[int, float]]]]:
        return self.transform.fill


class RandomPerspectiveMapper(TransformMapper[T.RandomPerspective]):
    source_type = T.RandomPerspective

    def __init__(self, transform: T.RandomPerspective):
        super().__init__(transform)

    @staticmethod
    def compute_homography(
        startpoints: List[Tuple[float, float]], endpoints: List[Tuple[float, float]]
    ) -> np.ndarray:
        assert len(startpoints) == len(endpoints) == 4

        a_matrix = torch.zeros(2 * len(startpoints), 8, dtype=torch.float)

        for i, (p1, p2) in enumerate(zip(endpoints, startpoints)):
            a_matrix[2 * i, :] = torch.tensor(
                [p1[0], p1[1], 1, 0, 0, 0, -p2[0] * p1[0], -p2[0] * p1[1]]
            )
            a_matrix[2 * i + 1, :] = torch.tensor(
                [0, 0, 0, p1[0], p1[1], 1, -p2[1] * p1[0], -p2[1] * p1[1]]
            )

        b_matrix = torch.tensor(startpoints, dtype=torch.float).view(8)
        res = torch.linalg.lstsq(a_matrix, b_matrix, driver="gels").solution

        m = np.eye(3, dtype=np.float32)
        m[0, :] = res[:3]
        m[1, :] = res[3:6]
        m[2, :2] = res[6:]

        return m

    def apply_transform(
        self, matrix: np.ndarray, dst_size: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, Any]:
        assert self.transform.fill == 0, "Only zero fill is supported"
        startpoints = None
        endpoints = None
        if torch.rand(1) <= self.transform.p:
            startpoints, endpoints = self.transform.get_params(
                dst_size[1], dst_size[0], self.transform.distortion_scale
            )
            # print(
            #     f"Perspective ds={self.transform.distortion_scale}: sp={startpoints} -> ep={endpoints}"
            # )
            matrix = self.compute_homography(endpoints, startpoints) @ matrix
        return matrix, dst_size, (self.source_type.__name__, startpoints, endpoints)

    def fill(
        self,
    ) -> Optional[Union[int, float, Tuple[Union[int, float], ...], List[Union[int, float]]]]:
        return self.transform.fill

    def interpolation(self) -> Optional[T.InterpolationMode]:
        return self.transform.interpolation


class CenterCropMapper(TransformMapper[T.CenterCrop]):
    source_type = T.CenterCrop

    def __init__(self, transform: T.CenterCrop):
        super().__init__(transform)

    def apply_transform(
        self, matrix: np.ndarray, dst_size: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, Any]:
        if isinstance(self.transform.size, numbers.Number):
            th, tw = int(self.transform.size), int(self.transform.size)
        elif isinstance(self.transform.size, (tuple, list)) and len(self.transform.size) == 1:
            th, tw = self.transform.size[0], self.transform.size[0]
        else:
            th, tw = self.transform.size

        shift_y = round(float(th - dst_size[0]) / 2)
        shift_x = round(float(tw - dst_size[1]) / 2)

        matrix = self.translate(shift_x, shift_y) @ matrix
        dst_size = np.array((th, tw), dtype=dst_size.dtype)
        # print(f"CenterCrop t=[{dx}, {dy}], s={dst_size}")
        return matrix, dst_size, (self.source_type.__name__, (shift_y, shift_x, th, tw))



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from types import TracebackType
from typing import BinaryIO, Optional, Type

from megatron.energon.epathlib import EPath


class RetryReadStream(BinaryIO):
    """A stream that retries reading from a file. Only supports reading bytes."""

    _path: EPath
    _file: Optional[BinaryIO]
    _pos: int
    _size: int

    def __init__(self, path: EPath):
        """Construct a RetryReadStream. It reads only bytes from a file."""
        self._path = path
        self._file = None
        self._pos = 0
        self._size = path.size()

    def __enter__(self) -> "RetryReadStream":
        return self

    def __exit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType
    ) -> None:
        self.close()

    def close(self) -> None:
        if self._file is not None:
            self._file.close()

    def read(self, n: int = -1) -> bytes:
        buf = b""
        for retry in range(10):
            try:
                if self._file is None:
                    self._file = self._path.open("rb")
                    self._file.seek(self._pos)
                res = self._file.read(n)
                self._pos += len(res)
                buf += res
                if (
                    (n == -1 and self._pos >= self._size)
                    or len(buf) == n
                    or self._pos >= self._size
                ):
                    return res
            except IOError:
                try:
                    self._file.close()
                except IOError:
                    pass
                self._file = None
                if retry == 9:
                    raise
                continue

    def seek(self, offset: int, whence: int = 0) -> int:
        if whence == 0:
            pass
        elif whence == 1:
            offset += self._pos
        elif whence == 2:
            offset += self._size
        else:
            raise ValueError(f"Invalid whence value: {whence}")
        offset = min(max(offset, 0), self._size)
        self._pos = offset
        try:
            if self._file is not None:
                self._file.seek(offset)
        except IOError:
            pass
        return self._pos

    def tell(self) -> int:
        return self._pos

    def isatty(self) -> bool:
        return False

    def readable(self) -> bool:
        return True

    def seekable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import inspect
import json
import re
import typing
from types import FunctionType
from typing import Any, List, Optional, Tuple, Type

import click
import yaml

from megatron.energon.epathlib import EPath
from megatron.energon.flavors import BaseWebdatasetFactory, CrudeWebdataset
from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.webdataset import MAIN_FOLDER_NAME
from megatron.energon.flavors.webdataset.metadata import (
    check_dataset_info_present,
    get_dataset_info,
)
from megatron.energon.metadataset.loader import prepare_metadataset


def type_str(tp: Type) -> str:
    """Returns a human-readable string for a type."""
    if typing.get_origin(tp) is not None:
        return repr(tp)
    if isinstance(tp, type):
        if tp.__module__ == "builtins":
            return tp.__qualname__
        return f"{tp.__module__}.{tp.__qualname__}"
    if tp is ...:
        return "..."
    if isinstance(tp, FunctionType):
        return tp.__name__
    return repr(tp)


def sample_loader_template(fields: dict, parts: list):
    """Returns a template for a sample_loader.py file."""

    fields_str = ""
    for field in fields:
        if field.name in ("__key__", "__restore_key__", "__subflavors__"):
            continue
        line = f"""        {field.name}=raw["TODO"],  # expected type: {type_str(field.type)}"""
        if field.default is not dataclasses.MISSING:
            line += ", default: " + repr(field.default)
        fields_str += line + "\n"

    return "\n".join(
        [
            "# This file was automatically generated by `energon prepare`.",
            "# TODO: Edit it to return the proper fields",
            "# import torch",
            "",
            "def sample_loader(raw: dict) -> dict:"
            "    # Note: Images are already decoded to tensors",
            "    # TODO: Set the correct values for all (required) fields",
            "    return dict(",
            fields_str,
            "    )",
            "",
            "def part_filter(part: str) -> bool:",
            "    # TODO: Filter for parts required by the sample_loader",
            "    # E.g. if your dataset contains jpeg, txt and json, but you won't use json,",
            "    # remove it from the list, such that it is not decoded. If you need all, keep as is",
            f"    return part in {tuple(parts)!r}",
            "",
        ]
    )


def printify_json(data: Any) -> Any:
    """Shortens json data to a human-readable length."""
    if isinstance(data, dict):
        return {k: printify_json(v) for k, v in data.items()}
    elif isinstance(data, list):
        if len(data) > 3:
            return [printify_json(v) for v in data[:3]] + ["..."]
        return [printify_json(v) for v in data]
    elif isinstance(data, str):
        return data[:25] + ("..." if len(data) > 25 else "")
    return data


@click.command(name="prepare")
@click.argument(
    "path",
    type=click.Path(path_type=EPath),
)
@click.option(
    "--progress/--no-progress",
    default=True,
)
@click.option(
    "--split-parts",
    help="Path pattern for parts in the form 'train:train/{000000-009999}.tar'. Will ignore ratio.",
    multiple=True,
    default=None,
)
@click.option(
    "--exclude",
    help="Exclude tar file paths (relative to root) matching that regex (at any position)",
)
@click.option(
    "--num-workers",
    type=int,
    default=16,
    help="Number of workers to use to index tar files",
)
@click.option(
    "--tar-index-only",
    help="Only (re)generate the tar-index",
    is_flag=True,
)
@click.option(
    "--shuffle-tars",
    help="If set, the tar files will be shuffled before splitting.",
    is_flag=True,
)
def command(
    path: EPath,
    progress: bool,
    split_parts: Optional[List[str]],
    exclude: str,
    num_workers: int,
    tar_index_only: bool,
    shuffle_tars: bool,
):
    """Prepare WebDataset for use with energon.

    The PATH should point to the folder with the dataset.
    This tool will add the required metadata yaml files to the dataset. See README.md for more
    details.
    """

    if path.is_file() and path.name.endswith(".yaml"):
        prepare_metadataset(path)
        return

    if tar_index_only:
        info = get_dataset_info(path)
        all_tars = list(info["shard_counts"].keys())
    else:
        if check_dataset_info_present(path):
            if not click.confirm(
                "It seems the dataset had already been prepared. Do you want to continue?"
            ):
                return

        all_tars = list(path.glob("**/*.tar")) + list(path.glob("**/*.tgz"))
        all_tars = [str(p.relative_to(path)) for p in sorted(all_tars)]

    if exclude:
        all_tars = [p for p in all_tars if not re.search(exclude, p)]

    if len(all_tars) == 0:
        click.echo("Did not find any tar files. Exiting.")
        return

    if not tar_index_only:
        click.echo(f"Found {len(all_tars)} tar files in total. The first and last ones are:")
        click.echo(f"- {all_tars[0]}")
        click.echo(f"- {all_tars[-1]}")
        click.echo(
            "If you want to exclude some of them, cancel with ctrl+c and specify an exclude "
            "filter in the command line."
        )

    split_parts_patterns: Optional[List[Tuple[str, str]]]
    if split_parts:
        split_parts_patterns = [tuple(x.split(":", 1)) for x in split_parts]
        split_parts_ratio = None
    elif not tar_index_only:
        split_input = click.prompt(
            'Please enter a desired train/val/test split like "0.5, 0.2, 0.3" or "8,1,1"', type=str
        )
        # Extract split floats
        try:
            split = [float(x.strip()) for x in split_input.split(",")]
            assert len(split) == 3
        except (ValueError, AssertionError):
            click.echo("Invalid split. Stopping.")
            return
        split_parts_ratio = [("train", split[0]), ("val", split[1]), ("test", split[2])]
        split_parts_patterns = None
    else:
        split_parts_ratio = None
        split_parts_patterns = None

    if progress:

        def progress_fn(els, length=None):
            with click.progressbar(
                els,
                label="Indexing shards",
                show_pos=True,
                length=length,
            ) as bar:
                yield from bar

    else:

        def progress_fn(els, length=None):
            return els

    found_types, duplicates = BaseWebdatasetFactory.prepare_dataset(
        path,
        all_tars,
        split_parts_ratio=split_parts_ratio,
        split_parts_patterns=split_parts_patterns,
        progress_fn=progress_fn,
        tar_index_only=tar_index_only,
        shuffle_seed=42 if shuffle_tars else None,
        workers=num_workers,
    )

    if duplicates:
        print(f"Examples of duplicates found: {duplicates}")
        print()
        print(
            "The dataset has duplicate keys. Best practice is to use unique keys. "
            "You won't be able to use this dataset for joining "
            "later on."
        )

    found_types = list(found_types)
    if tar_index_only:
        return

    if duplicates:
        if not click.confirm("Do you want to continue?"):
            return

    # Print json of first two samples
    for sample_idx, data in enumerate(
        BaseWebdatasetFactory.iter_dataset_content(path / all_tars[0], ("json",))
    ):
        print(f"Sample {sample_idx}, keys:")
        for key in data.keys():
            print(f" - {key}")
        if "json" in data:
            print(f"Json content of sample {sample_idx} of {all_tars[0]}:")
            print(json.dumps(printify_json(json.loads(data["json"])), indent=2))
        if sample_idx >= 1:
            break

    if len(found_types) > 10:
        click.echo(
            f"Found the following part types in the dataset: {', '.join(found_types[:10])} and more.."
        )
        allow_interactive_field_map = False
    else:
        click.echo(f"Found the following part types in the dataset: {', '.join(found_types)}")
        allow_interactive_field_map = True

    if click.confirm("Do you want to create a dataset.yaml interactively?", default=True):
        # Get a list of all classes in megatron.energon that are subclasses of WebdatasetBase
        import megatron.energon as data_import

        display_name_and_class = [
            (name, cls)
            for name, cls in inspect.getmembers(data_import)
            if isinstance(cls, type) and issubclass(cls, Sample)
        ]
        display_name_and_class.append(("Crude sample (plain dict for cooking)", CrudeWebdataset))

        # Print all classes and ask user to pick one
        click.echo("The following sample types are available:")
        for i, (name, cls) in enumerate(display_name_and_class):
            click.echo(f"{i}. {name}")
        while True:
            choice = click.prompt("Please enter a number to choose a class", type=int)
            try:
                _, cls = display_name_and_class[choice]
                break
            except IndexError:
                click.echo("Invalid choice. Please try again.")
                continue

        if cls == CrudeWebdataset:
            click.echo(
                "CrudeWebdataset does not need a field map. You will need to provide a `Cooker` for your dataset samples in your `TaskEncoder`."
            )
            click.echo(
                "Furthermore, you might want to add `subflavors` in your meta dataset specification."
            )
            dataset_definition = {
                "__module__": "megatron.energon",
                "__class__": cls.__name__,
            }
        else:
            click.echo("The sample type you selected:\n")
            click.echo(inspect.getsource(cls))

            dataset_definition = {
                "sample_type": {
                    "__module__": "megatron.energon",
                    "__class__": cls.__name__,
                },
            }

            if not allow_interactive_field_map:
                click.echo(
                    "You cannot set a field_map for this dataset. You will need a sample_loader."
                )

            if allow_interactive_field_map and click.confirm(
                "Do you want to set a simple field_map[Y] (or write your own sample_loader [n])?",
                default=True,
            ):
                click.echo(
                    "\nFor each field, please specify the corresponding name in the WebDataset."
                )
                click.echo(f"Available types in WebDataset: {', '.join(found_types)}")
                click.echo("Leave empty for skipping optional field")
                click.echo(
                    "You may also access json fields e.g. by setting the field to: json[field][field]"
                )
                click.echo("You may also specify alternative fields e.g. by setting to: jpg,png")

                click.echo(f"Please enter the field_map for {cls.__name__}:")

                dataset_definition["field_map"] = field_map = {}
                for field in dataclasses.fields(cls):
                    if field.name in (
                        "__key__",
                        "__restore_key__",
                        "__subflavors__",
                        "__sources__",
                    ):
                        continue
                    while True:
                        if (
                            field.default is dataclasses.MISSING
                            and field.default_factory is dataclasses.MISSING
                        ):
                            default = ""
                        elif field.default is not dataclasses.MISSING:
                            default = f", default: {field.default}"
                        elif field.default_factory is not dataclasses.MISSING:
                            default = f", default: {field.default_factory!r}"
                        else:
                            raise RuntimeError("This should never happen")
                        field_map[field.name] = input(
                            f"Please enter a webdataset field name for '{field.name}' "
                            f"({field.type}{default}): ",
                        )
                        if not field_map[field.name] and default:
                            del field_map[field.name]
                            break
                        type_ok = True
                        for option in field_map[field.name].split(","):
                            field_name = option.split("[", 1)[0]
                            if field_name not in found_types:
                                click.echo(
                                    "That type doesn't exist in the WebDataset. Please try again."
                                )
                                type_ok = False
                        if type_ok:
                            break
            else:
                if not allow_interactive_field_map:
                    template_part_types = set(["TODO"])
                else:
                    template_part_types = found_types

                if not (path / MAIN_FOLDER_NAME / "sample_loader.py").is_file() or click.confirm(
                    "Do you want to override the existing sample loader?"
                ):
                    with (path / MAIN_FOLDER_NAME / "sample_loader.py").open("w") as f:
                        f.write(
                            sample_loader_template(
                                dataclasses.fields(cls),
                                parts=template_part_types,
                            )
                        )
                    click.echo(
                        f"\nCreated {path / MAIN_FOLDER_NAME / 'sample_loader.py'}. Please edit it to "
                        f"return the proper values."
                    )
                dataset_definition["sample_loader"] = "sample_loader.py:sample_loader"
                dataset_definition["part_filter"] = "sample_loader.py:part_filter"

        # Write the dataset.yaml file
        with (path / MAIN_FOLDER_NAME / "dataset.yaml").open("w") as f:
            yaml.dump(dataset_definition, f, sort_keys=False)
    else:
        click.echo("You will have to add a dataset.yaml manually.")

    click.echo("Done")


if __name__ == "__main__":
    command()



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from megatron.energon.metadataset.dataset_loader import DatasetLoader
from megatron.energon.metadataset.loader import load_dataset, prepare_metadataset
from megatron.energon.metadataset.loader_interface import DatasetLoaderInterface
from megatron.energon.metadataset.metadataset import Metadataset
from megatron.energon.metadataset.metadataset_v2 import MetadatasetV2

__all__ = [
    "DatasetLoader",
    "load_dataset",
    "DatasetLoaderInterface",
    "Metadataset",
    "MetadatasetV2",
    "prepare_metadataset",
]



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import List, Type

import numpy as np
import PIL.Image
import torch
from torchvision import transforms as T
from torchvision.transforms.functional import InterpolationMode, pil_modes_mapping

from megatron.energon.transforms.custom import CustomTransform
from megatron.energon.transforms.mappers import TransformMapper


class MergeTransform(torch.nn.Module):
    transforms: List[TransformMapper]

    def __init__(self, transforms: List[torch.nn.Module]):
        super().__init__()
        self.transforms = [self._transform_to_mapper(t) for t in transforms]
        # Get common interpolation and fill values
        interpolation = None
        interpolation_t = None
        fill = None
        fill_t = None
        for t in self.transforms:
            t_fill = t.fill()
            if t_fill is not None:
                if fill is None:
                    fill = t_fill
                    fill_t = t
                if fill is not None and t_fill != fill:
                    raise ValueError(
                        f"Fill values are not equal: {fill} ({fill_t}) != {t_fill} ({t})"
                    )
            t_interpolation = t.interpolation()
            if t_interpolation is not None:
                if interpolation is None:
                    interpolation = t_interpolation
                    interpolation_t = t
                if interpolation is not None and t_interpolation != interpolation:
                    raise ValueError(
                        f"Interpolation values are not equal: {interpolation} ({interpolation_t}) != {t_interpolation} ({t})"
                    )

        self.interpolation = InterpolationMode.BILINEAR if interpolation is None else interpolation
        self.fill_value = fill

    def _transform_to_mapper(self, transform: torch.nn.Module) -> Type[TransformMapper]:
        """Given a transform object, instantiate the corresponding mapper.
        This also handles objects of derived transform classes."""

        if isinstance(transform, CustomTransform):
            # Custom transforms can be used as-is, they provide the apply_transform method
            return transform

        for m in TransformMapper.__subclasses__():
            if isinstance(transform, m.source_type):
                return m(transform)  # Instantiate
        raise ValueError(f"Unsupported transform type {type(transform)}")

    def forward(self, x):
        matrix = np.eye(3, dtype=np.float64)
        if isinstance(x, PIL.Image.Image):
            dst_size = np.array((x.height, x.width), dtype=np.int64)
        else:
            dst_size = np.array(x.shape[-2:], dtype=np.int64)
        all_params = []
        for transform in self.transforms:
            matrix, dst_size, params = transform.apply_transform(matrix, dst_size)
            all_params.append(params)

        if isinstance(x, PIL.Image.Image):
            try:
                interpolation = pil_modes_mapping[self.interpolation]
            except KeyError:
                raise NotImplementedError(f"interpolation: {self.interpolation}")

            # Invert matrix for backward mapping
            matrix = np.linalg.inv(matrix)

            # Scale matrix
            matrix /= matrix[2, 2]

            if self.fill_value is None:
                fill_color = None
            elif isinstance(self.fill_value, (int, float)):
                fill_color = (self.fill_value,) * len(x.getbands())
            else:
                fill_color = self.fill_value
            if np.allclose(matrix[2, :2], [0, 0]):
                # print("PIL Affine")
                return x.transform(
                    tuple(dst_size[::-1]),
                    PIL.Image.AFFINE,
                    matrix.flatten()[:6],
                    interpolation,
                    fillcolor=fill_color,
                )
            else:
                # print("PIL Perspective")
                return x.transform(
                    tuple(dst_size[::-1]),
                    PIL.Image.PERSPECTIVE,
                    matrix.flatten()[:8],
                    interpolation,
                    fillcolor=fill_color,
                )
        elif isinstance(x, torch.Tensor):
            print("torch affine")
            if self.interpolation == T.InterpolationMode.NEAREST:
                interpolation = "nearest"
            elif self.interpolation == T.InterpolationMode.BILINEAR:
                interpolation = "bilinear"
            elif self.interpolation == T.InterpolationMode.BICUBIC:
                interpolation = "bicubic"
            else:
                raise NotImplementedError(f"interpolation: {self.interpolation}")
            if self.fill_value is not None and self.fill_value != 0:
                raise NotImplementedError(
                    f"Fill value {self.fill_value} is not supported for torch"
                )
            # Normalize to [-1, 1] range
            matrix = (
                TransformMapper.translate(-1, -1)
                @ TransformMapper.scale(2 / dst_size[1], 2 / dst_size[0])
                @ matrix
                @ TransformMapper.scale(x.shape[-1] / 2, x.shape[-2] / 2)
                @ TransformMapper.translate(1, 1)
            )

            matrix = np.linalg.inv(matrix)
            if np.allclose(matrix[2, :2], [0, 0]):
                grid = torch.nn.functional.affine_grid(
                    torch.as_tensor(matrix[None, :2, :], dtype=torch.float32),
                    torch.Size((1, 3, *dst_size)),
                )
            else:
                xs = torch.linspace(-1, 1, dst_size[1], dtype=torch.float32)
                ys = torch.linspace(-1, 1, dst_size[0], dtype=torch.float32)
                zs = torch.ones((1,), dtype=torch.float32)
                # shape: (2<x,y,1>, W, H)
                grid = torch.stack(torch.meshgrid([xs, ys, zs], indexing="ij"))[..., 0]
                # shape: (H, W, 2<x,y,1>)
                grid = grid.permute(2, 1, 0)
                # shape: (H, W, 3<x,y,w>, 1)
                grid = (
                    torch.as_tensor(matrix, dtype=torch.float32)[None, None, ...] @ grid[..., None]
                )
                # shape: (H, W, 2<x,y>)
                grid = grid[:, :, :2, 0] / grid[:, :, 2:3, 0]
                # shape: (1, H, W, 2<x,y>)
                grid = grid[None, ...]
            return torch.nn.functional.grid_sample(
                x[None, ...], grid, interpolation, padding_mode="zeros", align_corners=False
            )[0, ...]
        else:
            raise NotImplementedError()
            # TODO: Needs implementation and testing
            import cv2

            return cv2.warpAffine(x, matrix[:2], tuple(dst_size), flags=cv2.INTER_LINEAR)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Callable, Dict, List, Optional, Union

from megatron.energon.epathlib import EPath
from megatron.energon.flavors.webdataset import DefaultDecoderWebdatasetFactory


class CrudeSample(dict):
    """Generic sample type to be processed later."""


class CrudeWebdataset(DefaultDecoderWebdatasetFactory[CrudeSample]):
    """The CrudeWebdataset is used to load crude / raw samples and
    decode them in the user code using so-called cookers.

    See the documentation under "Crude Data" for more information.
    """

    __sample_type__ = CrudeSample

    def __init__(
        self,
        path: EPath,
        *,
        subflavors: Optional[Dict[str, Any]] = None,
        part_filter: Union[str, List[str], Callable[[str], bool]] = lambda _: True,
        **kwargs,
    ):
        """
        Constructs a crude webdataset.

        Args:
            path: Root path to the joined datasets.
            subflavors: Subflavors dictionary to set for all loaded samples.
            part_filter: Function for filtering tar files to load by dict keys.
            **kwargs: Additional arguments to the BaseWebdataset constructor.
        """
        # We skip the parent class __init__ and call the BaseWebdataset.__init__ directly

        if "sample_loader" in kwargs:
            raise ValueError("sample_loader is not allowed to be set when using CrudeWebdataset")

        super().__init__(
            path,
            subflavors=subflavors,
            sample_loader=lambda sample: sample,
            part_filter=part_filter,
            **kwargs,
        )



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import functools
import inspect
from abc import ABC
from dataclasses import is_dataclass
from types import MethodType
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Hashable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import torch
from typing_extensions import ParamSpec

from megatron.energon.cache import CachePool, DecodeFileStore, FileStore
from megatron.energon.edataclass import edataclass
from megatron.energon.flavors import (
    CrudeSample,
    CrudeWebdataset,
    PinMemoryMixin,
    Sample,
    SampleDecoder,
    SavableDataset,
)
from megatron.energon.flavors.base_dataset import ExtendableDataclassMixin
from megatron.energon.metadataset.loader_interface import DatasetBlendMode, LoadedDataset
from megatron.energon.rng import SystemRng
from megatron.energon.source_info import SourceInfo
from megatron.energon.task_encoder.cooking import Cooker
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers import (
    BatchDataset,
    BlendDataset,
    ConcatDataset,
    EpochizeDataset,
    GroupBatchDataset,
    LimitDataset,
    LogSampleDataset,
    MapDataset,
    PackingDataset,
    ShuffleBufferDataset,
)
from megatron.energon.wrappers.repeat_dataset import RepeatDataset

T = TypeVar("T")
V = TypeVar("V")
T_sample = TypeVar("T_sample", bound=Sample)
T_encoded_sample = TypeVar("T_encoded_sample")
T_raw_batch = TypeVar("T_raw_batch")
T_batch = TypeVar("T_batch")


FeatureBatcher = Callable[[List[Any]], Any]


def generic_batch(batch: List[Any]) -> Any:
    """Based on the types/shapes of the batch: Will either pad and stack, or return as list.
    Recurses structures (dict, dataclass, namedtuple) and applies the same logic to each field."""
    if isinstance(batch[0], torch.Tensor):
        return batch_pad_stack(batch)
    elif isinstance(batch[0], dict):
        return {key: generic_batch([sample[key] for sample in batch]) for key in batch[0].keys()}
    elif is_dataclass(batch[0]):
        if hasattr(batch[0], "from_samples"):
            # The dataclass defines a method for batching samples
            return batch[0].from_samples(batch)
        return type(batch[0])(
            **{
                field.name: generic_batch([getattr(sample, field.name) for sample in batch])
                for field in dataclasses.fields(batch[0])
            }
        )
    elif isinstance(batch[0], tuple) and hasattr(batch[0], "_fields"):
        # NamedTuple
        return type(batch[0])(
            **{
                field: generic_batch([getattr(sample, field) for sample in batch])
                for field in batch[0]._fields
            }
        )
    else:
        return batch_list(batch)


def batch_stack(batch: List[Any]) -> Any:
    """Stack a batch of tensors."""
    return torch.stack(batch, dim=0)


def batch_pad_stack(batch: List[Any]) -> Any:
    """Stack a batch of arbitrary-sized tensors padded with 0s."""
    max_size = [max(b.shape[dim] for b in batch) for dim in range(batch[0].ndim)]
    batch_tensor = batch[0].new_zeros((len(batch), *max_size))
    for i, b in enumerate(batch):
        batch_tensor[(i, *(slice(0, s) for s in b.shape))] = b
    # Pad all tensors to max_size
    return batch_tensor


def batch_list(batch: List[Any]) -> Any:
    """Stack a batch of tensors padded with 0s."""
    return batch


P = ParamSpec("P")


@overload
def stateless(*, restore_seeds: bool = False) -> Callable[[Callable[P, T]], Callable[P, T]]: ...


@overload
def stateless(fn: Callable[P, T]) -> Callable[P, T]: ...


def stateless(
    fn: Optional[Callable[..., T]] = None,
    *,
    restore_seeds: bool = False,
    failure_tolerance: Optional[int] = None,
) -> Union[Callable[[Callable[..., T]], Callable[..., T]], Callable[..., T]]:
    """Decorator to mark a function of the task encoder as restorable.

    Args:
        fn: The function to decorate.
        restore_seeds: Whether to restore the seeds for the function. I.e. the seeds are set
            from the sample index and the worker seed, such that they can be restored when a sample
            is restored from that function.
        failure_tolerance: The number of consecutive exceptions that are handled, after which a `FatalSampleError` is
            raised for this function.

    Usage:

    .. code-block:: python

        @stateless
        def encode_sample(self, sample: T_sample) -> T_encoded_sample:
            ...

        # Or if randomness is used (e.g. for augmentations):
        @stateless(restore_seeds=True)
        def encode_sample(self, sample: T_sample) -> T_encoded_sample:
            ...

    """

    if fn is None:
        return lambda f: stateless(
            f, restore_seeds=restore_seeds, failure_tolerance=failure_tolerance
        )
    if restore_seeds:
        worker_seed = None

        @functools.wraps(fn)
        def seed_wrapper_generator(self, *args, **kwargs):
            nonlocal worker_seed
            if worker_seed is None:
                worker_seed = WorkerConfig.active_worker_config.worker_seed()

            # Save the RNG states and set the new seed
            outer_rng_state = SystemRng.save_state()

            # Before constructing the generator and before the first
            # iteration, set inner RNG based on seed computed
            # from worker_seed and current sample index
            SystemRng.seed_args(worker_seed, self.current_sample_index)

            it = iter(fn(self, *args, **kwargs))

            inner_rand_state = None

            while True:
                if inner_rand_state is not None:
                    # Restore inner random state before calling the generator
                    # This will not be done on the first iteration
                    SystemRng.restore_state(inner_rand_state)

                try:
                    # Now call the generator. This will yield the sample
                    # But note it may also throw an exception or a StopIteration
                    sample = next(it)

                    # Save inner random state after calling the generator
                    inner_rand_state = SystemRng.save_state()
                except StopIteration:
                    # We're stopping here, but the outer random state
                    # will be restored before returning (in finally below)
                    break
                finally:
                    # Restore outer rand state before yielding or when an exception was raised
                    SystemRng.restore_state(outer_rng_state)

                # Now yield the sample.
                # This will give control back to the caller who may
                # change the random state.
                yield sample

                # Save outer random state after yielding
                outer_rng_state = SystemRng.save_state()

        @functools.wraps(fn)
        def seed_wrapper(self, *args, **kwargs):
            nonlocal worker_seed
            if worker_seed is None:
                worker_seed = WorkerConfig.active_worker_config.worker_seed()

            # Save the RNG states and set the new seed
            rng_state = SystemRng.save_state()

            SystemRng.seed_args(worker_seed, self.current_sample_index)

            try:
                return fn(self, *args, **kwargs)
            finally:
                # Restore the RNGs
                SystemRng.restore_state(rng_state)

        if inspect.isgeneratorfunction(fn):
            setattr(seed_wrapper_generator, "__stateless__", True)
            return seed_wrapper_generator
        else:
            setattr(seed_wrapper, "__stateless__", True)
            return seed_wrapper

    setattr(fn, "__stateless__", True)
    setattr(fn, "__failure_tolerance__", failure_tolerance)
    return fn


def get_stateless(fn: Callable) -> bool:
    """Get whether a function is stateless."""
    return getattr(fn, "__stateless__", False)


def get_failure_tolerance(
    fn: Callable, default_failure_tolerance: Optional[int] = None
) -> Optional[int]:
    """Get the failure tolerance of a function."""
    return getattr(fn, "__failure_tolerance__", default_failure_tolerance)


@edataclass
class Batch(PinMemoryMixin, ExtendableDataclassMixin):
    """Base class for a batch dataclass. Provides a default implementation for pinning memory.
    Additionally, it provides a future safe implementation for creating an instance from another
    batch `Batch.derive_from`."""

    #: Uniquely identifies each sample in the dataset.
    __key__: list[str]
    #: Key for restoring the sample. This is used to restore the sample from a checkpoint. It
    # should be a (nested) tuple of strings and integers, which can be used to index the dataset.
    __restore_key__: Tuple[Union[str, int, tuple], ...]

    #: A dataset may define a subflavors to distinguish between samples of the same sample type.
    __subflavors__: Optional[list[Optional[Dict[str, Any]]]] = None

    #: Information about the source of the sample, i.e. where the data was loaded from.
    __sources__: Optional[tuple[SourceInfo, ...]] = None

    @classmethod
    def derive_from(cls: Type[T_batch], base_batch: "Batch", **kwargs) -> T_batch:
        """
        Uses the base fields of `Batch` from base_batch (i.e. __key__, __restore_key__, __subflavors__, __sources__)
        and creates a new batch with the kwargs as fields. This is useful for creating new batches, while keeping the
        metadata of the base batch.

        Use like::

        .. code-block:: python

            def encode_batch(batch: RawBatch) -> Batch:
                return Batch.derive_from(batch, field1=batch.field1 + 1)

        Args:
            base_batch: The base batch to copy the base fields / metadata from.
            kwargs: The fields of the new batch.

        Returns:
            The new batch.
        """
        base_kwargs = {
            field.name: getattr(base_batch, field.name) for field in dataclasses.fields(Batch)
        }
        return cls(
            **base_kwargs,
            **kwargs,
        )

    @classmethod
    def from_samples(cls: Type[T_batch], samples: Sequence[Sample], **kwargs) -> T_batch:
        """
        Creates a batch from samples to be batched. Tensors will be padded and stacked, other types will be put into
        lists. This is the default implementation for `Batch.from_samples`.

        Args:
            samples: The samples to batch.
            kwargs: Additional (overriding) fields of the batch.

        Returns:
            The constructed batch.
        """
        assert all(dataclasses.is_dataclass(scls) for scls in samples), (
            "Samples must be dataclasses"
        )
        # assert dataclasses.is_dataclass(cls), "Batch must be dataclass"
        init_args = {}
        fields = dataclasses.fields(cls)
        for field in fields:
            if field.name in kwargs:
                init_args[field.name] = kwargs[field.name]
            elif field.name == "__sources__":
                if any(sample.__sources__ is not None for sample in samples):
                    # Special handling, needs flattening
                    init_args[field.name] = tuple(
                        source
                        for sample in samples
                        if sample.__sources__
                        for source in sample.__sources__
                    )
            elif field.name == "__subflavors__":
                if any(sample.__subflavors__ is not None for sample in samples):
                    init_args[field.name] = [
                        sample.__subflavors__ for sample in samples if sample.__subflavors__
                    ]
            else:
                value = [getattr(sample, field.name) for sample in samples]
                if len(samples) > 0 and isinstance(samples[0], torch.Tensor):
                    value = batch_pad_stack(value)
                init_args[field.name] = value
        return cls(**init_args)


class TaskEncoder(ABC, Generic[T_sample, T_encoded_sample, T_raw_batch, T_batch]):
    """
    Base class for task encoders.

    Task encoding follows these steps:
      0. Data comes from the dataset
      1. :meth:`megatron.energon.TaskEncoder.encode_sample` / :meth:`megatron.energon.TaskEncoder.preencode_sample` is called on each sample
      2. :meth:`megatron.energon.TaskEncoder.select_samples_to_pack` is called on the buffer of samples
      3. :meth:`megatron.energon.TaskEncoder.postencode_sample` is called on each sample of the current pack
      4. :meth:`megatron.energon.TaskEncoder.pack_selected_samples` is called on the selected sample pack
      5. :meth:`megatron.energon.TaskEncoder.batch` is called on the list of encoded samples
      6. :meth:`megatron.energon.TaskEncoder.encode_batch` is called on the batch
      7. yield to main process
      8. :meth:`megatron.energon.Batch.to_device` is called on the encoded batch
      9. resulting encoded batch is passed to the network
    """

    __default_failure_tolerance__: Optional[int] = 100

    cookers: Sequence[Cooker[T_sample]] = ()
    #: Internal: List of registered cookers. Will be the same as `cookers` after registering cookers.
    _registered_cookers: List[Cooker[T_sample]]

    #: The decoder to use for decoding samples. Set manually as needed to override options.
    decoder: Optional[SampleDecoder] = SampleDecoder()

    @stateless
    def cook_crude_sample(
        self,
        sample: Union[T_sample, CrudeSample],
        get_primary_aux: Callable[[], FileStore],
        **aux: FileStore,
    ) -> T_sample:
        """
        Cooks a crude sample.

        Args:
            sample: The sample to cook.
            get_primary_aux: A function that returns the (cached) primary auxiliary dataset.
            **aux: The auxiliary side dishes to use for cooking.

        Returns: The cooked sample.
        """
        if isinstance(sample, CrudeSample):
            for cooker in self.cookers:
                if cooker.is_match(sample):
                    assert get_stateless(cooker.cook), "Cooker must be stateless"
                    if not cooker.need_primary and not cooker.need_cache:
                        kwargs = aux
                    else:
                        kwargs: dict = {}
                        if cooker.need_primary:
                            kwargs["primary"] = get_primary_aux()
                        kwargs.update(aux)
                        if cooker.need_cache:
                            kwargs["cache"] = self.cache
                    return cooker.cook(sample, **kwargs)

            raise NotImplementedError(
                "You are using crude samples but not providing a way to cook them: "
                f"Sample key={sample['__key__']}, subflavors={sample['__subflavors__']}, "
                f"self.cookers={self.cookers}"
            )
        else:
            assert isinstance(sample, Sample), "Sample must be a complete Sample or a CrudeSample"
            return sample

    def _is_overridden(
        self, bound_method: Callable[..., Any], bases: Optional[Sequence[Type[Any]]] = None
    ) -> bool:
        """Check if a method is overridden by a subclass of the base class(es).
        By default, only TaskEncoder is used as a base class.
        This is mainly used for optimization purposes. If the default method
        is a no-op, we can skip it entirely unless the user has overridden it.

        Args:
            bound_method: The method to check.
            bases: The base classes to check against.
        Returns:
            True if the method is overridden outside of TaskEncoder, False otherwise.
        """

        if not isinstance(bound_method, MethodType):
            # If the method is not bound, it is always overridden
            return True

        # Get the underlying function
        func = bound_method.__func__

        # Check if the subclass method matches any of the base class methods
        if bases is None:
            bases = (TaskEncoder,)
        return not any(getattr(base, func.__name__) is func for base in bases)

    @stateless
    def encode_sample(
        self, sample: T_sample
    ) -> Union[T_encoded_sample, Generator[T_encoded_sample, None, None]]:
        """Encode a single sample. May raise :exc:`megatron.energon.SkipSample` to skip a sample.
        Alternatively, this can be a generator that yields (or ignores) new samples.
        If this is defined, :func:`preencode_sample` and :func:`postencode_sample` must not be defined.
        """
        return sample

    @stateless
    def preencode_sample(
        self, sample: T_sample
    ) -> Union[T_sample, Generator[T_sample, None, None]]:
        """Pre-encode a single sample. May raise :exc:`megatron.energon.SkipSample` to skip a sample.
        Alternatively, this can be a generator that yields (or ignores) new samples.
        Use in conjunction with packing and caching.
        If this is defined, :func:`encode_sample` must not be defined.
        """
        return sample

    @stateless
    def postencode_sample(
        self, sample: T_sample
    ) -> Union[T_encoded_sample, Generator[T_encoded_sample, None, None]]:
        """Post-encode a single sample. May raise :exc:`megatron.energon.SkipSample` to skip a sample.
        Alternatively, this can be a generator that yields (or ignores) new samples.
        Use in conjunction with packing and caching.
        If this is defined, :func:`encode_sample` must not be defined.
        """
        return sample

    @stateless
    def batch(self, samples: List[T_encoded_sample]) -> T_raw_batch:
        """Move a batch to a device. May raise :exc:`megatron.energon.SkipSample` to skip a batch."""
        return self._batch(samples, type(samples[0]))

    def batch_group_criterion(self, sample: T_encoded_sample) -> Tuple[Hashable, Optional[int]]:
        """
        Return a group criterion for the sample. Default implementation does not group
        (effectively, it returns a single value `(None, None)`, thus only one group is used).
        Returns the key of the bucket to put this sample into, and the size of the bucket (=batch size).
        The bucket size must always be the same for the same bucket key.

        May raise :exc:`megatron.energon.SkipSample` to skip a batch.
        """
        return None, None

    @stateless
    def encode_batch(self, batch: T_raw_batch) -> Union[T_batch, Generator[T_batch, None, None]]:
        """Encode a batch of samples. May raise :exc:`megatron.energon.SkipSample` to skip a batch.
        Alternatively, this can be a generator that yields (or ignores) new batches."""
        return batch

    def _batch(
        self,
        samples: List[T_encoded_sample],
        result_type: Type[T_raw_batch],
        actions: Optional[Dict[str, FeatureBatcher]] = None,
        default_action: FeatureBatcher = generic_batch,
    ) -> T_raw_batch:
        """
        Batch a list of samples.

        Args:
            samples: The samples to batch
            result_type: Type of the result (might be dict, dataclass, or namedtuple)
            actions: For each field (=key), may specify a specific batcher
            default_action: The batcher to apply to all fields not in `action`

        Returns:
            The batched result
        """
        if dataclasses.is_dataclass(result_type) and hasattr(result_type, "from_samples"):
            return result_type.from_samples(samples)

        # Get dict of samples
        if isinstance(samples[0], dict):
            list_samples = {key: [sample[key] for sample in samples] for key in samples[0].keys()}
        elif is_dataclass(samples[0]):
            list_samples = {
                field.name: [getattr(sample, field.name) for sample in samples]
                for field in dataclasses.fields(samples[0])
            }
        elif isinstance(samples[0], tuple) and hasattr(samples[0], "_fields"):
            # NamedTuple
            list_samples = {
                field: [getattr(sample, field) for sample in samples]
                for field in samples[0]._fields
            }
        else:
            raise ValueError("Unrecognized sample type.")
        # Convert each field
        if actions is not None:
            list_samples = {
                key: default_action(value) if key not in actions else actions[key](value)
                for key, value in list_samples.items()
            }
        else:
            list_samples = {key: default_action(value) for key, value in list_samples.items()}
        # Construct result
        if issubclass(result_type, dict):
            return list_samples
        elif dataclasses.is_dataclass(result_type) or issubclass(result_type, tuple):
            # DataClass or NamedTuple
            return result_type(**list_samples)
        else:
            raise ValueError("Unrecognized result type.")

    def select_samples_to_pack(
        self, samples: List[T_encoded_sample]
    ) -> List[List[T_encoded_sample]]:
        """
        For packing, selects the samples to be packed together.
        Packing is only active when packing_buffer_size is set.
        Internally this stage is called "pre_packing".

        Args:
            samples: The samples to pre-pack. A full buffer will be passed into the function.

        Returns: The pre-packed samples as a list of lists of samples.
        """
        raise NotImplementedError("Packing only effective when overridden.")

    def pack_selected_samples(self, samples: List[T_encoded_sample]) -> T_encoded_sample:
        """
        Given one set of samples to pack, returns the final packed sample.
        Packing is only active when packing_buffer_size is set.
        Internally this stage is called "final_packing".

        Args:
            samples: The samples to pack into a single sample

        Returns: The final packed sample.
        """
        raise NotImplementedError("Packing only effective when overridden.")

    def build_batch(
        self,
        dataset: SavableDataset[T_encoded_sample],
        *,
        batch_size: Optional[int],
        batch_drop_last: bool = False,
        packing_buffer_size: Optional[int] = None,
        worker_config: WorkerConfig,
    ) -> SavableDataset[T_raw_batch]:
        """Applies the batcher to the dataset."""

        dataset: SavableDataset[Any]

        if packing_buffer_size is not None:
            select_samples_to_pack_provided = self._is_overridden(self.select_samples_to_pack)
            pack_selected_samples_provided = self._is_overridden(self.pack_selected_samples)

            assert select_samples_to_pack_provided and pack_selected_samples_provided, (
                "Both select_samples_to_pack and pack_selected_samples methods must be provided in the TaskEncoder when using packing_buffer_size"
            )

            if self._is_overridden(self.postencode_sample):
                post_encode_fn = self.postencode_sample
            else:
                post_encode_fn = None

            dataset = PackingDataset(
                dataset,
                buffer_size=packing_buffer_size,
                pre_packer=self.select_samples_to_pack,
                final_packer=self.pack_selected_samples,
                final_packer_stateless=get_stateless(self.pack_selected_samples),
                sample_encoder=post_encode_fn,
                sample_encoder_stateless=True
                if post_encode_fn is None
                else get_stateless(post_encode_fn),
                worker_config=worker_config,
                pre_packer_failure_tolerance=get_failure_tolerance(
                    self.select_samples_to_pack, self.__default_failure_tolerance__
                ),
                final_packer_failure_tolerance=get_failure_tolerance(
                    self.pack_selected_samples, self.__default_failure_tolerance__
                ),
                sample_encoder_failure_tolerance=None
                if post_encode_fn is None
                else get_failure_tolerance(post_encode_fn, self.__default_failure_tolerance__),
            )
        elif self._is_overridden(self.postencode_sample):
            dataset = MapDataset(
                dataset,
                self.postencode_sample,
                worker_config=worker_config,
                stateless_map_fn=get_stateless(self.postencode_sample),
                failure_tolerance=get_failure_tolerance(
                    self.postencode_sample, self.__default_failure_tolerance__
                ),
            )

        if self._is_overridden(self.batch_group_criterion):
            dataset = GroupBatchDataset(
                dataset,
                fixed_batch_size=batch_size,
                sample_group_key=self.batch_group_criterion,
                batcher=self.batch,
                drop_last=batch_drop_last,
                worker_config=worker_config,
                failure_tolerance=get_failure_tolerance(
                    self.batch, self.__default_failure_tolerance__
                ),
            )

            if self._is_overridden(self.encode_batch):
                dataset = MapDataset(
                    dataset,
                    self.encode_batch,
                    worker_config=worker_config,
                    stateless_map_fn=get_stateless(self.encode_batch),
                    failure_tolerance=get_failure_tolerance(
                        self.encode_batch, self.__default_failure_tolerance__
                    ),
                )
        else:
            # No grouping is active

            if batch_size is not None:
                dataset = BatchDataset(
                    dataset,
                    batch_size=batch_size,
                    batcher=self.batch,
                    batcher_stateless=get_stateless(self.batch),
                    drop_last=batch_drop_last,
                    worker_config=worker_config,
                    failure_tolerance=get_failure_tolerance(
                        self.batch, self.__default_failure_tolerance__
                    ),
                )

                if self._is_overridden(self.encode_batch):
                    dataset = MapDataset(
                        dataset,
                        self.encode_batch,
                        worker_config=worker_config,
                        stateless_map_fn=get_stateless(self.encode_batch),
                        failure_tolerance=get_failure_tolerance(
                            self.encode_batch, self.__default_failure_tolerance__
                        ),
                    )

        return dataset

    def build_cook_crude_sample(
        self,
        dataset: SavableDataset[Union[T_sample, dict]],
        *,
        worker_config: WorkerConfig,
        subflavors: Dict[str, Any],
        get_primary_aux: Callable[[], FileStore],
        aux: Optional[Dict[str, FileStore]] = None,
    ) -> SavableDataset[T_sample]:
        """Applies the sample cooker to the dataset if we have cookers registered."""

        assert self.cookers, "No cookers registered, but got crude dataset."

        if aux is not None and self.decoder is not None:
            aux = {k: DecodeFileStore(v, decoder=self.decoder) for k, v in aux.items()}

        # Cache the primary auxiliary dataset for this dataset, i.e. construct it once when needed
        primary_aux = None

        def _get_primary_aux():
            nonlocal primary_aux
            if primary_aux is None:
                try:
                    if aux is not None:
                        primary_aux = aux.get("primary")
                    if primary_aux is None:
                        primary_aux = get_primary_aux()
                    assert primary_aux is not None, "Primary auxiliary dataset must always exist"
                    if self.decoder is not None:
                        primary_aux = DecodeFileStore(primary_aux, decoder=self.decoder)
                except Exception as e:
                    # Make the exception throw through for the sample being loaded
                    raise SystemError("Error getting primary auxiliary dataset") from e
            return primary_aux

        if aux is not None:
            cook_fn = functools.partial(
                self.cook_crude_sample, get_primary_aux=_get_primary_aux, **aux
            )
        else:
            cook_fn = functools.partial(self.cook_crude_sample, get_primary_aux=_get_primary_aux)

        return MapDataset(
            dataset,
            cook_fn,
            worker_config=worker_config,
            stateless_map_fn=True,
            map_fn_config=dict(
                cookers=[
                    dict(
                        cook=SavableDataset._function_config(cooker.cook),
                        has_subflavors=cooker.has_subflavors,
                    )
                    for cooker in self.cookers
                ],
                subflavors=subflavors,
            ),
            failure_tolerance=get_failure_tolerance(cook_fn, self.__default_failure_tolerance__),
        )

    def _load_dataset(
        self, dataset: LoadedDataset, worker_rotation_offset: int, worker_config: WorkerConfig
    ) -> SavableDataset[T_sample]:
        """Loads a train dataset, optionally cooking the samples."""
        if dataset.dataset.__sample_type__ == CrudeSample:
            return self.build_cook_crude_sample(
                dataset.dataset.build(worker_rotation_offset=worker_rotation_offset),
                worker_config=worker_config,
                subflavors=dataset.dataset.subflavors,
                get_primary_aux=dataset.dataset.as_file_store,
                aux=dataset.aux,
            )
        else:
            assert dataset.aux is None, "Aux is not supported for non-crude datasets."
            return dataset.dataset.build(worker_rotation_offset=worker_rotation_offset)

    def build_encode_sample(
        self,
        dataset: SavableDataset[T_sample],
        *,
        worker_config: WorkerConfig,
    ) -> SavableDataset[T_encoded_sample]:
        """Applies the sample encoder to the dataset."""
        if self._is_overridden(self.preencode_sample):
            pre_encode_fn = self.preencode_sample
            assert not self._is_overridden(
                self.encode_sample, bases=(TaskEncoder, DefaultTaskEncoder)
            ), "Cannot have both pre- and post-encode functions defined."
        elif self._is_overridden(self.encode_sample):
            pre_encode_fn = self.encode_sample
        else:
            pre_encode_fn = None
        if pre_encode_fn is not None:
            dataset = MapDataset(
                dataset,
                pre_encode_fn,
                worker_config=worker_config,
                stateless_map_fn=get_stateless(pre_encode_fn),
                failure_tolerance=get_failure_tolerance(
                    pre_encode_fn, self.__default_failure_tolerance__
                ),
            )
        return dataset

    def build_train_datasets(
        self,
        *,
        datasets: List[LoadedDataset],
        worker_config: WorkerConfig,
        batch_size: Optional[int],
        batch_drop_last: bool = False,
        packing_buffer_size: Optional[int] = None,
        virtual_epoch_length: int = 0,
        shuffle_buffer_size: Optional[int] = None,
        blend_mode: DatasetBlendMode = DatasetBlendMode.NONE,
        repeat: bool = True,
    ) -> SavableDataset[T_batch]:
        """Combines train datasets to a single dataset."""

        # Check if there's a CrudeWebdataset but no cookers
        for dataset in datasets:
            if isinstance(dataset.dataset, CrudeWebdataset):
                assert self.cookers, "CrudeWebdataset found, but no cookers registered."

        global_workers = max(1, worker_config.num_workers) * worker_config.world_size
        rotation_lengths = [len(dataset.dataset) for dataset in datasets]
        for i in range(1, len(rotation_lengths)):
            rotation_lengths[i] += rotation_lengths[i - 1]
        worker_rotation_offsets = [
            rotation_length % global_workers for rotation_length in [0] + rotation_lengths[:-1]
        ]

        if blend_mode == DatasetBlendMode.DATASET_WEIGHT:
            assert repeat, (
                "If repeat is False, the datasets can only be repeated or have no mode. Cannot blend with dataset weights."
            )
            inner_datasets = [
                (
                    RepeatDataset(
                        self._load_dataset(
                            dataset, worker_rotation_offset, worker_config=worker_config
                        ),
                        worker_config=worker_config,
                    ),
                    1.0 if dataset.weight is None else float(dataset.weight),
                )
                for dataset, worker_rotation_offset in zip(datasets, worker_rotation_offsets)
            ]
            # Already repeating the inner datasets, so no need to repeat again
            repeat = False
        elif blend_mode == DatasetBlendMode.SAMPLE_REPETITIONS or (
            not repeat and blend_mode == DatasetBlendMode.NONE
        ):
            inner_datasets = [
                (
                    (
                        self._load_dataset(
                            dataset, worker_rotation_offset, worker_config=worker_config
                        )
                        if dataset.repetitions is None or dataset.repetitions == 1
                        else RepeatDataset(
                            self._load_dataset(
                                dataset, worker_rotation_offset, worker_config=worker_config
                            ),
                            repeats=dataset.repetitions,
                            worker_config=worker_config,
                        )
                    ),
                    len(dataset.dataset)
                    * (1 if dataset.repetitions is None else dataset.repetitions),
                )
                for dataset, worker_rotation_offset in zip(datasets, worker_rotation_offsets)
            ]
        else:
            inner_datasets = [
                (
                    RepeatDataset(
                        self._load_dataset(
                            dataset, worker_rotation_offset, worker_config=worker_config
                        ),
                        worker_config=worker_config,
                    ),
                    1.0,
                )
                for dataset, worker_rotation_offset in zip(datasets, worker_rotation_offsets)
            ]
            # Already repeating the inner datasets, so no need to repeat again
            repeat = False

        if len(inner_datasets) > 1:
            # The worker offset for each dataset is the cumsum of the dataset lengths, but modulo the
            # global number of workers.
            dataset = BlendDataset(
                *[inner_dataset[:2] for inner_dataset in inner_datasets],
                worker_config=worker_config,
            )
        elif len(datasets) == 1:
            dataset = inner_datasets[0][0]
        else:
            raise ValueError("No datasets given.")
        if repeat:
            # Still need to repeat the dataset
            dataset = RepeatDataset(dataset, worker_config=worker_config)
        if shuffle_buffer_size is not None and shuffle_buffer_size > 1:
            dataset = ShuffleBufferDataset(
                dataset,
                size=shuffle_buffer_size,
                worker_config=worker_config,
            )
        dataset = self.build_encode_sample(dataset, worker_config=worker_config)
        dataset = self.build_batch(
            dataset,
            batch_size=batch_size,
            batch_drop_last=batch_drop_last,
            packing_buffer_size=packing_buffer_size,
            worker_config=worker_config,
        )
        if virtual_epoch_length > 0:
            dataset = EpochizeDataset(
                dataset,
                length=virtual_epoch_length,
                worker_config=worker_config,
            )
        if worker_config.should_log(level=1):
            dataset = LogSampleDataset(dataset, mode="train", worker_config=worker_config)

        return dataset

    def build_val_datasets(
        self,
        *,
        datasets: List[LoadedDataset],
        worker_config: WorkerConfig,
        batch_size: int,
        batch_drop_last: bool = False,
        packing_buffer_size: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> SavableDataset[T_batch]:
        """Combines val datasets to a single dataset."""

        # Check if there's a CrudeWebdataset but no cookers
        for dataset in datasets:
            if isinstance(dataset, CrudeWebdataset):
                assert self.cookers, "CrudeWebdataset found, but no cookers registered."

        global_workers = max(1, worker_config.num_workers) * worker_config.world_size
        rotation_lengths = [len(dataset.dataset) for dataset in datasets]
        for i in range(1, len(rotation_lengths)):
            rotation_lengths[i] += rotation_lengths[i - 1]
        worker_rotation_offsets = [
            rotation_length % global_workers for rotation_length in [0] + rotation_lengths[:-1]
        ]

        if len(datasets) > 1:
            dataset = ConcatDataset(
                *[
                    self._load_dataset(dataset, worker_rotation_offset, worker_config)
                    for dataset, worker_rotation_offset in zip(datasets, worker_rotation_offsets)
                ],
                worker_config=worker_config,
            )
        elif len(datasets) == 1:
            dataset = self._load_dataset(datasets[0], worker_rotation_offsets[0], worker_config)
        else:
            raise ValueError("No datasets given.")
        dataset = self.build_encode_sample(dataset, worker_config=worker_config)
        dataset = self.build_batch(
            dataset,
            batch_size=batch_size,
            batch_drop_last=batch_drop_last,
            packing_buffer_size=packing_buffer_size,
            worker_config=worker_config,
        )
        if limit is not None and limit > 0:
            dataset = LimitDataset(
                dataset,
                length=limit,
                worker_config=worker_config,
                reset_after_epoch=True,
            )

        if worker_config.should_log(level=2):
            dataset = LogSampleDataset(dataset, mode="val", worker_config=worker_config)

        return dataset

    @property
    def current_batch_index(self) -> int:
        """Returns the current index for the next batch yielded from the current worker. Each batch
        on the current rank will get a strictly increasing unique number. Counting happens on each
        rank separately (i.e. each rank will get the same numbers for same batch index)."""
        assert WorkerConfig.active_worker_config is not None, (
            "The batch_index can only be fetched within the worker, and to be usable, you must use the get_(savable_)loader methods provided from the package."
        )
        return WorkerConfig.active_worker_config.active_worker_batch_index

    @property
    def current_sample_index(self) -> int:
        """Returns the current index for the next sample yielded from the current routine (e.g.
        for `encode_sample`, `batch`, or `encode_batch`). Each routine will get a number
        representing the number of calls to that function. Across workers, this number will be
        unique, but it is not synced across workers, thus it may raise in different intervals (e.g.
        if batching does not work the same for all batches). When restoring a sample, this number is
        also restored and can be relied on for deterministic randomness reproduction of a sample."""
        assert WorkerConfig.active_worker_config is not None, (
            "The batch_index can only be fetched within the worker, and to be usable, you must use the get_(savable_)loader methods provided from the package."
        )
        return WorkerConfig.active_worker_config.active_worker_sample_index

    @property
    def cache(self) -> CachePool:
        """Returns the cache pool to use for caching out sample data to disk (for use with cookers / aux file stores).
        This is set and configured externally by the loader."""
        assert WorkerConfig.active_worker_config is not None, (
            "The cache can only be fetched within the worker, and to be usable, you must use the get_(savable_)loader methods provided from the package."
        )
        assert WorkerConfig.active_worker_config._cache_pool is not None, (
            "Cache pool must be set by the loader."
        )
        return WorkerConfig.active_worker_config._cache_pool


class DefaultTaskEncoder(
    TaskEncoder[T_sample, T_encoded_sample, T_raw_batch, T_batch],
    ABC,
    Generic[T_sample, T_encoded_sample, T_raw_batch, T_batch],
):
    """The default task encoder supports automagically mapping to target types.
    You may override any methods to customize the behavior. By default, `encode_sample` is the
    identity function, `batch` calls `\\_batch` with the type of the first sample, and `encode\\_batch`
    is also the identity function. If you set any of `encoded_sample_type`, `raw_batch_type` or
    `batch_type`, the corresponding method return that type, where it automatically maps the fields
    (by name) to your new type.
    """

    _encoded_sample_type: Optional[Type[T_encoded_sample]]
    _raw_batch_type: Optional[Type[T_raw_batch]]
    _batch_type: Optional[Type[T_batch]]

    def __init__(
        self,
        *,
        encoded_sample_type: Optional[Type[T_encoded_sample]] = None,
        raw_batch_type: Optional[Type[T_raw_batch]] = None,
        batch_type: Optional[Type[T_batch]] = None,
    ):
        """
        Initialize the default task encoder.
        Types may be:
          * A `@dataclass` class: Return that typed dataclass. Field names must match the input
            fields.
          * A `NamedTuple` class: Return that typed namedtuple. Field names must match the input
            fields.
          * `dict`: Simply return the input as dict with field names as keys.

        Args:
            encoded_sample_type: Type of encoded samples (before batching)
            raw_batch_type: Type of the batched samples (after batching)
            batch_type: Type of the encoded batched samples
            cache: Cache pool to use for caching. If not provided, a no-op cache pool will be used.
        """
        self._encoded_sample_type = encoded_sample_type
        self._raw_batch_type = raw_batch_type
        self._batch_type = batch_type

    @stateless
    def encode_sample(
        self, sample: T_sample
    ) -> Union[T_encoded_sample, Generator[T_encoded_sample, None, None]]:
        """Encode a single sample. The default implementation converts to the
        _encoded_sample_type."""
        if self._encoded_sample_type is None or isinstance(sample, self._encoded_sample_type):
            return sample
        if is_dataclass(sample):
            fields = {
                field.name: getattr(sample, field.name) for field in dataclasses.fields(sample)
            }
        elif isinstance(sample, tuple) and hasattr(sample, "_fields"):
            fields = {field: getattr(sample, field) for field in sample._fields}
        elif isinstance(sample, dict):
            fields = sample
        else:
            raise ValueError("Unrecognized sample type.")
        if issubclass(self._encoded_sample_type, dict):
            return fields
        elif dataclasses.is_dataclass(self._encoded_sample_type) or issubclass(
            self._encoded_sample_type, tuple
        ):
            # DataClass or NamedTuple
            return self._encoded_sample_type(**fields)
        else:
            raise ValueError("Unrecognized encoded sample type.")

    @stateless
    def batch(self, samples: List[T_encoded_sample]) -> T_raw_batch:
        """Batch a list of samples. The default implementation uses default batching to convert
        to _batch_type."""
        actions = None
        if isinstance(samples[0], Sample):
            actions = {
                "__subflavors__": lambda x: x,
            }
        return self._batch(
            samples,
            type(samples[0]) if self._raw_batch_type is None else self._raw_batch_type,
            actions=actions,
        )

    @stateless
    def encode_batch(self, batch: T_raw_batch) -> Union[T_batch, Generator[T_batch, None, None]]:
        """Encode a batch of samples. The default implementation converts to the
        _encoded_batch_type."""
        if self._batch_type is None or self._raw_batch_type == self._batch_type:
            return batch
        if is_dataclass(batch):
            fields = {field.name: getattr(batch, field.name) for field in dataclasses.fields(batch)}
        elif isinstance(batch, tuple) and hasattr(batch, "_fields"):
            fields = {field: getattr(batch, field) for field in batch._fields}
        elif isinstance(batch, dict):
            fields = batch
        else:
            raise ValueError("Unrecognized sample type.")
        if issubclass(self._batch_type, dict):
            return fields
        elif dataclasses.is_dataclass(self._batch_type) or issubclass(self._batch_type, tuple):
            # DataClass or NamedTuple
            return self._batch_type(**fields)
        else:
            raise ValueError("Unrecognized encoded sample type.")


class AugmentTaskEncoder(
    TaskEncoder[T_sample, T_encoded_sample, T_raw_batch, T_batch],
    Generic[T_sample, T_encoded_sample, T_raw_batch, T_batch],
):
    """Augment a task encoder with additional functionality. By default, delegates everything to the
    original task encoder."""

    def __init__(self, task_encoder: TaskEncoder[T_sample, T_encoded_sample, T_raw_batch, T_batch]):
        """Initialize the augmenting task encoder.

        Args:
            task_encoder: The delegate task encoder. All calls will by default be forwarded to this.
        """
        self._task_encoder = task_encoder

    @property
    def decoder(self) -> SampleDecoder:
        return self._task_encoder.decoder

    def encode_sample(self, sample: T_sample) -> T_encoded_sample:
        return self._task_encoder.encode_sample(sample)

    def batch(self, samples: List[T_encoded_sample]) -> T_raw_batch:
        return self._task_encoder.batch(samples)

    def encode_batch(self, batch_data: T_raw_batch) -> T_batch:
        return self._task_encoder.encode_batch(batch_data)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Optional, TypeVar

from megatron.energon.cache.base import CachePool, FileStore, Lazy
from megatron.energon.edataclass import edataclass
from megatron.energon.source_info import SourceInfo, add_source_info

T = TypeVar("T")


@edataclass
class DirectLazy(Lazy[T]):
    """
    This is not really lazy, it will just defer the dataset access to the first get().
    """

    _data: Optional[tuple[T, SourceInfo]] = None

    def get(self, sample: Any = None) -> T:
        """Get the lazy data now and adds no source info to the sample."""
        if self._data is None:
            self._data = self.ds[self.fname]
        assert self._data is not None
        add_source_info(sample, self._data[1])
        return self._data[0]

    def __hash__(self) -> int:
        """Allows usage in sets and dicts as key."""
        return hash((id(self.ds), self.fname))

    def __eq__(self, other: Any) -> bool:
        """Allows usage in sets and dicts as key. Compares the data source and the filename."""
        if not isinstance(other, Lazy):
            return False
        return self.ds is other.ds and self.fname == other.fname


class NoCachePool(CachePool):
    """A pass-through cache pool that does not cache anything."""

    def get(self, ds: FileStore, fname: str, sample: Any = None) -> Any:
        """Get the data for a given file and adds the source info to the sample."""
        return ds.get(fname, sample)

    def get_lazy(self, ds: FileStore, fname: str) -> DirectLazy:
        return DirectLazy(ds=ds, fname=fname, pool=self)

    def close(self) -> None:
        pass



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause


import hashlib
from pathlib import Path

from filelock import FileLock

from megatron.energon.epathlib import EPath

LOCAL_COPY_TMP_DIR = Path("/tmp/energon_local_copy")


def ensure_local_copy(path: EPath) -> EPath:
    """If the path is not local, copy it to a temporary directory and return the
    path to the temporary directory.

    Args:
        path: The path to the file to copy.

    Returns:
        The path to the local copy of the file or the original path if it is already local.
    """

    assert path.is_file(), f"Path {path} is not a file"

    if path.is_local():
        return path

    LOCAL_COPY_TMP_DIR.mkdir(parents=True, exist_ok=True)
    digest = hashlib.sha256(str(path).encode()).hexdigest()
    final_path = LOCAL_COPY_TMP_DIR / f"{digest}.bin"
    lock_path = final_path.with_suffix(".lock")
    tmp_path = final_path.with_suffix(".part")

    # Block until lock is free
    with FileLock(lock_path, timeout=60 * 5):
        if final_path.exists():  # someone else already produced it
            return EPath(final_path)

        # We are the downloader
        try:
            path.copy(EPath(tmp_path))
            tmp_path.rename(final_path)
        finally:
            tmp_path.unlink(missing_ok=True)

    return EPath(final_path)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import re

split_name_re = re.compile(r"^((?:.*/|)[^.]+)[.]([^/]*)$")
skip_meta_re = re.compile(r"__[^/]*__($|/)")
MAIN_FOLDER_NAME = ".nv-meta"
JOIN_INDEX_FILENAME = "join_index.bin"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, TypeVar

from megatron.energon.edataclass import edataclass
from megatron.energon.source_info import SourceInfo, add_source_info

T = TypeVar("T")


class FileStore(Generic[T]):
    """Base type for a dataset that can be accessed randomly by sample key."""

    @abstractmethod
    def __getitem__(self, key: str) -> tuple[T, SourceInfo]:
        """Returns the data for the given key."""
        ...

    def get(self, key: str, sample: Any = None) -> Any:
        """Returns the data for the given key and adds the source info to the sample."""
        data, source_info = self[key]
        add_source_info(sample, source_info)
        return data

    @abstractmethod
    def get_path(self) -> str:
        """Returns the path to the dataset."""
        ...


@edataclass
class Lazy(Generic[T]):
    """
    Abstract base class for lazy references to data.
    """

    ds: FileStore
    fname: str
    pool: "CachePool"

    @abstractmethod
    def get(self, sample: Any = None) -> T:
        """
        Get the lazy data now and adds the source info to the sample.
        """
        ...

    def __hash__(self) -> int:
        """Allows usage in sets and dicts as key."""
        return hash((id(self.ds), self.fname))

    def __eq__(self, other: Any) -> bool:
        """Allows usage in sets and dicts as key. Compares the data source and the filename."""
        if not isinstance(other, Lazy):
            return False
        return self.ds is other.ds and self.fname == other.fname


@edataclass
class MockLazy(Lazy[T]):
    """
    Mock object, which can be used as a Lazy. Allows the user to set the function to retrieve the
    data. May be used to create a Lazy that is initialized from a function.
    """

    ds: FileStore
    fname: str
    pool: "CachePool"

    get_fn: Callable[[str], T]

    def __init__(self, fname: str, get_fn: Callable[[str], T]):
        """
        Initialize the MockLazy object.

        Args:
            fname: The file name of the mock object (may be used by the user).
            get_fn: The function to retrieve/generate the data.
        """
        self.ds = None
        self.fname = fname
        self.pool = None
        self.get_fn = get_fn

    def get(self, sample: Any = None) -> T:
        """
        Get the lazy data now and adds no source info to the sample.
        """
        return self.get_fn(self.fname)

    def __hash__(self) -> int:
        return hash((self.fname, self.get_fn))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, MockLazy):
            return False
        return self.fname == other.fname and self.get_fn == other.get_fn

    def __repr__(self) -> str:
        return f"MockLazy(fname={self.fname!r}, get_fn={self.get_fn!r})"


class CachePool(ABC):
    """
    A cache pool allows to load the needed data in the background and access it later.
    The most important example being `FileStoreCachePool` which caches data on a local SSD disk.

    To request data, use `get_lazy` to get a `Lazy` object.
    Then, call `Lazy.get()` to get the data later on.
    """

    @abstractmethod
    def get(self, ds: FileStore, fname: str, sample: Any = None) -> Any:
        """
        Get the data for a given file and adds the source info to the sample.
        """
        ...

    @abstractmethod
    def get_lazy(self, ds: FileStore, fname: str) -> Lazy:
        """
        Get a lazy reference to the data for a given file.
        """
        ...

    @abstractmethod
    def close(self) -> None:
        """
        Close the cache pool.
        """
        ...


class FileStoreDecoder(ABC):
    """
    Abstract base class for decoders.
    """

    @abstractmethod
    def decode(self, fname: str, data: bytes) -> Any:
        """
        Decode the specified file (i.e. path/key.ext). The extension is used to select the decoder.

        Args:
            fname: The file name of the file to decode.
            raw: The raw bytes of the file to decode.

        Returns:
            The decoded field's data.
        """
        ...



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import av
from sortedcontainers import SortedList

from ..keyframeinfo import KeyframeInfo


def parse_probe(file):
    keyframes = {}
    with av.open(file) as input_container:
        for stream_idx, stream in enumerate(input_container.streams.video):
            packet_pts = [
                (index, p.pts)
                for index, p in enumerate(input_container.demux(video=stream_idx))
                if p.is_keyframe
            ]
            packet_pts.sort(key=lambda x: x[1])

            keyframes[stream.id] = SortedList([KeyframeInfo(*p) for p in packet_pts])

        return keyframes



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC, abstractmethod

from megatron.energon.state import FlexState


class Savable(ABC):
    @abstractmethod
    def save_state(self) -> FlexState:
        """Dumps the state of the object into some serializable format.
        Should only dump the part that is persistable for energon usage.
        """
        ...

    @abstractmethod
    def restore_state(self, state: FlexState):
        """Restores the state of the object.
        If state is None, the object is restored to its default state."""
        ...



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Generic, Type, TypeVar

from megatron.energon.epathlib import EPath
from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.webdataset.default_decoder_webdataset import (
    DefaultDecoderWebdatasetFactory,
)

T_sample = TypeVar("T_sample", covariant=True)


class StandardWebdatasetFactory(DefaultDecoderWebdatasetFactory[T_sample], Generic[T_sample]):
    """
    This dataset sample loader factory uses the sample type e.g. given from a dataset.yaml, and applies the default
    loading logic, which includes decoding images, videos and containers.
    """

    def __init__(
        self,
        path: EPath,
        *,
        sample_type: Type[T_sample],
        **kwargs,
    ):
        """
        Factory for the standard webdataset sample loader.

        Args:
            path: Path to the dataset (passed to parent)
            sample_type: Type of the sample to be loaded
            auto_decode: If true, use the default webdataset sample decoder.
            image_decode: This defines the decoding results.
            ignore_decoder_errors: If true, ignore errors when decoding.
            subflavors: Subflavors dictionary to set for all loaded samples.
            field_map: Mapping from the webdataset fields to the sample fields.
            sample_loader: Function to load the sample from the webdataset fields. May be a string
                in order to load a function from a module, or a callable directly.
            part_filter: Filter for the parts to load. May be a string in order to load a function
                from a module, or a callable directly.
            split_part: Which part to load (e.g. 'train', 'val', 'test').
            training: If true, apply shuffling and loop the dataset.
            worker_config: Configuration for the workers.
            shuffle_over_epochs: Only effective if training=True.
                How many epochs to shuffle over if training.
                If = 1, every sample is seen exactly once per epoch.
                If > 1, samples (or rather shard slices) are shuffled within this number of epochs
                (i.e. randomly selected without replacement).
                If -1, the shards are effectively shuffle over infinite epochs (i.e. shard slices
                are drawn with replacement).
            parallel_shard_iters: Number of parallel opened shards per worker, shuffling between.
            max_samples_per_sequence: Maximum number of samples per sequence (=how many samples
                    will be sequentially iterated).
            split_config: Config file to use for shard split definitions.
            handler: Exception handler. Args: (exception, key).
        """
        self.__sample_type__ = sample_type
        assert issubclass(sample_type, Sample), (
            f"sample_type must be subclass of megatron.energon.Sample, got {sample_type.__qualname__}"
        )
        super().__init__(path, **kwargs)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Dict, Generic, Iterator, TypeVar

from megatron.energon.flavors.base_dataset import SavableDataset, add_sample_restore_key
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.base import BaseWrapperDataset

T_sample = TypeVar("T_sample")


class ConcatDataset(BaseWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """
    This dataset wrapper concatenates multiple iterable datasets together. The datasets must be
    finite, otherwise not all datasets can be sampled. This is only useful for validation / test
    datasets.
    """

    def __init__(
        self,
        *datasets: SavableDataset[T_sample],
        worker_config: WorkerConfig,
    ):
        """Construct a concatenated dataset."""
        super().__init__(datasets, worker_config=worker_config)
        assert len(self) >= 0, "Datasets must be finite."

    def reset_state_own(self) -> None:
        return

    def __len__(self):
        return sum(len(dataset) for dataset in self.datasets)

    def __iter__(self) -> Iterator[T_sample]:
        for ds_idx, dataset in enumerate(self.datasets):
            for sample in dataset:
                yield add_sample_restore_key(
                    sample,
                    ds_idx,
                    src=self,
                )

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "datasets": [dataset.config() for dataset in self.datasets],
        }

    def __str__(self):
        return f"ConcatDataset(datasets={self.datasets})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import itertools
import warnings
from functools import wraps
from typing import Any, Type, TypeVar, Union

import numpy as np
import torch


def compact_str(
    value: Union[dict, list, str, int, bool, None],
    depth: int = 3,
    max_items: int = 10,
    max_str_len: int = 50,
) -> str:
    """
    Compact representation of a value as a string.

    Args:
        value: The value to compact
        depth: The maximum depth to compact
        max_items: The maximum number of items to show in a list or dict
        max_str_len: The maximum string length to show

    Returns: The printable string
    """
    if isinstance(value, dict):
        if depth <= 0:
            return "{...}"
        return (
            "{"
            + ", ".join(
                (
                    f"{k}: {v!r}"
                    if isinstance(k, str) and k.startswith("__")
                    else f"{k}: {compact_str(v, depth - 1, max_items, max_str_len)}"
                )
                for k, v in itertools.islice(value.items(), max_items)
            )
            + "}"
        )
    elif isinstance(value, list):
        if depth <= 0:
            return "[...]"
        return (
            "["
            + ", ".join(
                compact_str(v, depth - 1, max_items, max_str_len) for v in value[:max_items]
            )
            + "]"
        )
    elif isinstance(value, tuple):
        if depth <= 0:
            return "(...)"
        return (
            "("
            + ", ".join(
                compact_str(v, depth - 1, max_items, max_str_len) for v in value[:max_items]
            )
            + ")"
        )
    elif isinstance(value, str):
        if len(value) > max_str_len:
            return repr(value[:max_str_len] + "...")
        return repr(value)
    elif isinstance(value, torch.Tensor):
        return f"Tensor(shape={value.shape}, dtype={value.dtype}, device={value.device})"
    elif isinstance(value, np.ndarray):
        return f"np.ndarray(shape={value.shape}, dtype={value.dtype})"
    elif dataclasses.is_dataclass(value):
        return f"{value.__class__.__name__}({', '.join(f'{field.name}={compact_str(getattr(value, field.name))}' for field in dataclasses.fields(value))})"
    else:
        return compact_str(repr(value), depth, max_items, max_str_len)


T = TypeVar("T")


class SampleException(ValueError):
    @classmethod
    def from_sample_key(cls: Type[T], sample_key: str) -> T:
        return cls(f"Sample {sample_key} failed")

    @classmethod
    def from_sample(cls: Type[T], sample: Any, message: str = "") -> T:
        if message:
            message = f": {message}"
        return cls(f"Sample {compact_str(sample)} failed{message}")


class FatalSampleError(SampleException):
    # This will not be handled by the error handler
    pass


def warn_deprecated(reason, stacklevel=2):
    warnings.warn(reason, FutureWarning, stacklevel=stacklevel)


def deprecated(reason):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            warn_deprecated(f"{func.__name__} is deprecated: {reason}", stacklevel=3)
            return func(*args, **kwargs)

        return wrapper

    return decorator


SYSTEM_EXCEPTIONS = (
    SystemError,
    SyntaxError,
    ImportError,
    StopIteration,
    StopAsyncIteration,
    MemoryError,
    RecursionError,
    ReferenceError,
    NameError,
    UnboundLocalError,
    FatalSampleError,
)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from megatron.energon.errors import SYSTEM_EXCEPTIONS, FatalSampleError
from megatron.energon.flavors.base_dataset import SavableDataset, add_sample_restore_key
from megatron.energon.source_info import SourceInfo
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers._log_exception import log_exception
from megatron.energon.wrappers.base import BaseWrapperDataset, SampleIndex, get_sample_restore_key
from megatron.energon.wrappers.skip import SkipSample

T_sample = TypeVar("T_sample")
T_sample_out = TypeVar("T_sample_out")


class MapDataset(BaseWrapperDataset[T_sample, T_sample_out], Generic[T_sample, T_sample_out]):
    """This dataset wrapper applies a custom function to transform each sample."""

    map_fn: Callable[[T_sample], Union[T_sample_out, Generator[T_sample_out, None, None]]]
    error_handler: Callable[[Exception, T_sample, list[SourceInfo]], None]
    stateless_map_fn: bool
    map_fn_config: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]]
    _sample_index: SampleIndex
    _generator_sample_key: Optional[Any]
    _generator_offset: Optional[int]

    _savable_fields = (
        "_sample_index",
        "_generator_sample_key",
        "_generator_offset",
    )

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        map_fn: Callable[[T_sample], Union[T_sample_out, Generator[T_sample_out, None, None]]],
        *,
        error_handler: Callable[[Exception, T_sample, list[SourceInfo]], None] = log_exception,
        stateless_map_fn: bool = False,
        map_fn_config: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]] = None,
        failure_tolerance: Optional[int] = 100,
        worker_config: WorkerConfig,
    ):
        """Construct a MapDataset.

        If this should be savable, the map_fn must only return a sample, or a generator yielding
        0 or 1 sample per input sample. Otherwise this will be broken (see `IterMapDataset`).

        Args:
            dataset: The input dataset to wrap
            map_fn: The function to apply to each sample. May raise
                :exc:`megatron.energon.SkipSample` to skip a sample. Alternatively, may return a
                generator to yield multiple or no samples.
            error_handler: Handler for errors. Defaults to logging and ignoring the exception.
            stateless_map_fn: If true, the map_fn is deterministic and stateless
                (thus key for random access can propagate to inner dataset). Defaults to False.
            map_fn_config: Configuration for the map_fn function. If callable, it should return the
                configuration. Defaults to None.
            failure_tolerance: The number of consecutive failures after which the dataset is considered broken.
            worker_config: Worker configuration.
        """
        super().__init__(dataset, worker_config=worker_config)
        self.map_fn = map_fn
        self.error_handler = error_handler
        self.stateless_map_fn = stateless_map_fn
        self.map_fn_config = map_fn_config
        self.failure_tolerance = failure_tolerance

        self.reset_state_own()

    def reset_state_own(self) -> None:
        self._sample_index = SampleIndex(self.worker_config, src=self)
        self._generator_sample_key = None
        self._generator_offset = None

    def __len__(self):
        return len(self.dataset)

    def __iter__(self) -> Iterator[T_sample_out]:
        last_map_failures = 0

        if self._generator_sample_key is not None:
            assert self._generator_offset is not None
            sample = self.dataset.restore_sample(self._generator_sample_key)
            # Do not increment the sample index, use previous index
            with self._sample_index.ctx(self._sample_index.current_idx) as sample_idx:
                mapped_sample = self.map_fn(sample)
            assert isinstance(mapped_sample, Generator)
            assert inspect.isgeneratorfunction(self.map_fn), (
                f"Generator in {self.map_fn} but not marked as such."
            )
            target_offset = self._generator_offset
            self._generator_offset = 0
            for idx, (sample_idx, inner_sample) in enumerate(
                self._sample_index.iter_ctx(mapped_sample, sample_idx)
            ):
                # Skip other samples
                if idx >= target_offset:
                    self._generator_offset = idx + 1
                    yield add_sample_restore_key(
                        inner_sample,
                        sample_idx,
                        idx,
                        src=self,
                    )
            self._generator_sample_key = None
            self._generator_offset = None

        for sample in self.dataset:
            restore_key = get_sample_restore_key(sample)
            try:
                with self._sample_index.ctx() as sample_idx:
                    mapped_sample = self.map_fn(sample)
                if isinstance(mapped_sample, Generator):
                    assert inspect.isgeneratorfunction(self.map_fn), (
                        f"Generator in {self.map_fn} but not marked as such."
                    )
                    self._generator_sample_key = restore_key
                    self._generator_offset = 0
                    # In case of a generator, additionally store the index of the yielded samples
                    # per input sample
                    for idx, (sample_idx, inner_sample) in enumerate(
                        self._sample_index.iter_ctx(mapped_sample, sample_idx)
                    ):
                        self._generator_offset = idx + 1
                        last_map_failures = 0
                        yield add_sample_restore_key(
                            inner_sample,
                            sample_idx,
                            idx,
                            src=self,
                        )
                    self._generator_sample_key = None
                    self._generator_offset = None
                else:
                    last_map_failures = 0
                    yield add_sample_restore_key(
                        mapped_sample,
                        sample_idx,
                        src=self,
                    )
            except GeneratorExit:
                raise
            except SkipSample:
                pass
            except SYSTEM_EXCEPTIONS:
                raise FatalSampleError.from_sample(sample)
            except Exception as e:
                self.error_handler(e, sample)
                last_map_failures += 1
                if (
                    self.failure_tolerance is not None
                    and last_map_failures >= self.failure_tolerance
                ):
                    raise FatalSampleError.from_sample(
                        sample,
                        f"MapDataset {self.map_fn} failed {last_map_failures} times in a row. Likely your code or dataset are broken.",
                    )

    def can_restore_sample(self) -> bool:
        return super().can_restore_sample() and self.stateless_map_fn

    def assert_can_restore(self) -> None:
        assert self.stateless_map_fn, (
            f"MapDataset can only restore samples if map_fn {self.map_fn} is stateless."
        )
        super().assert_can_restore()

    def restore_sample(self, restore_key: Tuple[Union[str, int, tuple], ...]) -> T_sample_out:
        self.assert_can_restore()
        if inspect.isgeneratorfunction(self.map_fn):
            id, sample_idx, local_idx = restore_key[:3]
            assert id == type(self).__name__
            restore_key = restore_key[3:]
            assert isinstance(local_idx, int)
        else:
            id, sample_idx = restore_key[:2]
            assert id == type(self).__name__
            restore_key = restore_key[2:]
        inner_sample = self.dataset.restore_sample(restore_key)
        with self._sample_index.ctx(sample_idx):
            mapped_sample = self.map_fn(inner_sample)
        if isinstance(mapped_sample, Generator):
            assert inspect.isgeneratorfunction(self.map_fn), (
                f"Generator in {self.map_fn} but not marked as such."
            )
            for idx, (sample_idx, res_sample) in enumerate(
                self._sample_index.iter_ctx(mapped_sample, sample_idx)
            ):
                if idx == local_idx:
                    return add_sample_restore_key(res_sample, sample_idx, local_idx, src=self)
            assert False, (
                "Generator did not yield enough samples, but is marked stateless/deterministic."
            )
        else:
            return add_sample_restore_key(mapped_sample, sample_idx, src=self)

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "map_fn": self._function_config(self.map_fn),
            **(
                {
                    "map_fn_config": (
                        self.map_fn_config() if callable(self.map_fn_config) else self.map_fn_config
                    )
                }
                if self.map_fn_config
                else {}
            ),
            "map_fn_stateless": self.stateless_map_fn,
        }

    def __str__(self):
        return f"MapDataset(map_fn={self.map_fn}, dataset={self.dataset})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from megatron.energon.av.av_data import AVData
from megatron.energon.av.av_decoder import (
    AVDecoder,
    AVMetadata,
    AVWebdatasetDecoder,
)
from megatron.energon.av.utils import get_clips_uniform, get_single_frames_uniform

__all__ = [
    "AVDecoder",
    "AVMetadata",
    "AVWebdatasetDecoder",
    "AVData",
    "get_clips_uniform",
    "get_single_frames_uniform",
]



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import re
import shutil
from pathlib import Path as PathlibPath
from pathlib import PurePosixPath
from typing import BinaryIO, Generator, Literal, Optional, TextIO, Tuple, Union, overload

import multistorageclient as msc

from megatron.energon.errors import warn_deprecated


class EPath:
    """EPath work much like pathlib.Path but does support local as well as remote file systems.
    To initialize a remote path, prepend the path with "msc://" to use the Multi-Storage Client (MSC).
    For example:

        EPath("msc://profilename/my_datasets/webdataset-000.tar")

    You will need to have your MSC configuration (~/.msc_config.yaml) set up to access the object stores
    or use your rclone configuration. See https://nvidia.github.io/multi-storage-client/config/index.html
    for more information.
    """

    # The path without the protocol. Can also be in S3 for example
    internal_path: PurePosixPath
    # The profile used to access the file system
    profile: str
    # The file system
    fs: msc.StorageClient

    def __init__(
        self,
        initial_path: Union[str, "EPath", PathlibPath],
    ) -> None:
        if isinstance(initial_path, EPath):
            self.internal_path = initial_path.internal_path
            self.profile = initial_path.profile
            self.fs = initial_path.fs
        else:
            if isinstance(initial_path, PathlibPath):
                path = str(initial_path.absolute())
                profile = "default"
            else:
                protocol, profile, path = self._split_protocol(initial_path)
                if protocol is None or protocol == "file":
                    profile = "default"
                    path = str(PathlibPath(path).absolute())
                elif protocol == "rclone":
                    warn_deprecated("rclone:// protocol is deprecated. Use msc:// instead.")
                else:
                    assert protocol == "msc", f"Unknown protocol: {protocol}"
            if not path.startswith("/"):
                path = "/" + path
            self.internal_path = self._resolve(path)
            assert profile is not None
            self.profile = profile
            # Resolve the client. Only depends on the protocol and the first part of the path
            self.fs, _ = msc.resolve_storage_client(f"msc://{self.profile}")

    def __getstate__(self) -> dict:
        return {
            "internal_path": self.internal_path,
            "profile": self.profile,
            # Do not save the fs when serializing, to avoid leaking credentials
        }

    def __setstate__(self, state: dict) -> None:
        self.internal_path = state["internal_path"]
        self.profile = state["profile"]
        self.fs, _ = msc.resolve_storage_client(f"msc://{self.profile}")

    @staticmethod
    def _resolve(path: Union[str, PurePosixPath]) -> PurePosixPath:
        """Resolve a path, removing .. and . components."""
        if isinstance(path, str):
            path = PurePosixPath(path)
        parts = path.parts
        if parts[0] != "/":
            raise ValueError("Only absolute paths are supported")
        if ".." in parts or "." in parts:
            new_parts = []
            for part in parts[1:]:
                if part == "..":
                    if len(new_parts) == 0:
                        raise ValueError(f"Path above root: {path}")
                    new_parts.pop()
                elif part == ".":
                    pass
                else:
                    new_parts.append(part)
            path = PurePosixPath("/", *new_parts)
        return path

    @staticmethod
    def _split_protocol(path: str) -> Tuple[Optional[str], Optional[str], str]:
        regex = re.compile(r"^(?P<protocol>[a-z]+)://(?P<profile>[^/]+?)/(?P<path>.+)$")
        m = regex.match(path)
        if m is None:
            return None, None, path
        return m.group("protocol"), m.group("profile"), m.group("path")

    @property
    def _internal_str_path(self) -> str:
        """Return the path as used inside the file system, without the protocol and fs part."""
        return str(self.internal_path)

    @overload
    def open(self, mode: Literal["r", "w"] = "r", block_size: Optional[int] = None) -> TextIO: ...

    @overload
    def open(self, mode: Literal["rb", "wb"], block_size: Optional[int] = None) -> BinaryIO: ...

    def open(
        self, mode: Literal["r", "rb", "w", "wb"] = "r", block_size: Optional[int] = None
    ) -> Union[TextIO, BinaryIO]:
        return self.fs.open(self._internal_str_path, mode)

    def read_text(self) -> str:
        with self.open() as f:
            return f.read()

    def read_bytes(self) -> bytes:
        with self.open("rb") as f:
            return f.read()

    def write_text(self, text: str) -> None:
        with self.open("w") as f:
            f.write(text)

    def write_bytes(self, data: bytes) -> None:
        with self.open("wb") as f:
            f.write(data)

    def copy(self, target: "EPath") -> None:
        """Copy a file to a new path, possibly between different file systems.

        Args:
            target: The path to the local file to download to.
        """

        if self.is_file():
            if self.fs == target.fs:
                self.fs.copy(self._internal_str_path, target._internal_str_path)
            elif target.is_local():
                self.fs.download_file(self._internal_str_path, target._internal_str_path)
            elif self.is_local():
                target.fs.upload_file(target._internal_str_path, self._internal_str_path)
            else:
                with self.open("rb") as src_f, target.open("wb") as dst_f:
                    shutil.copyfileobj(src_f, dst_f)
        else:
            inner_path = EPath(self)
            for fpath in self.fs.list(self._internal_str_path):
                inner_path.internal_path = PurePosixPath("/" + fpath.key)
                inner_path.copy(target / inner_path.relative_to(self))

    @property
    def name(self) -> str:
        return self.internal_path.name

    @property
    def parent(self) -> "EPath":
        new_path = EPath(self)
        new_path.internal_path = self.internal_path.parent
        return new_path

    @property
    def url(self) -> str:
        if self.is_local():
            return self._internal_str_path
        int_path_str = str(self.internal_path)
        return f"msc://{self.profile}{int_path_str}"

    def is_local(self) -> bool:
        return self.profile == "default"

    def is_dir(self) -> bool:
        return self.fs.info(self._internal_str_path).type == "directory"

    def is_file(self) -> bool:
        return self.fs.is_file(self._internal_str_path)

    def mkdir(self, exist_ok: bool = True, parents: bool = False):
        pass

    def glob(self, pattern) -> Generator["EPath", None, None]:
        search_path_pattern = (self / pattern)._internal_str_path

        for path in self.fs.glob(search_path_pattern):
            assert isinstance(path, str)

            new_path = EPath(self)
            new_path.internal_path = self._resolve(self.internal_path / PurePosixPath(path))

            yield new_path

    def size(self) -> int:
        return self.fs.info(self._internal_str_path).content_length

    def with_suffix(self, suffix: str) -> "EPath":
        new_path = EPath(self)
        new_path.internal_path = self.internal_path.with_suffix(suffix)
        return new_path

    def move(self, target: "EPath") -> None:
        self.copy(target)
        self.unlink()

    def unlink(self) -> None:
        return self.fs.delete(self._internal_str_path)

    def relative_to(self, other: "EPath") -> str:
        assert self.profile == other.profile, "Can only use relative_to within same profile"

        return str(self.internal_path.relative_to(other.internal_path))

    def __truediv__(self, other: Union[str, "EPath"]) -> "EPath":
        if isinstance(other, EPath):
            # Always absolute
            return other
        if other.startswith("/") or "://" in other:
            return EPath(other)

        new_path = EPath(self)
        new_path.internal_path = self._resolve(self.internal_path / other)
        return new_path

    def __lt__(self, other: "EPath") -> bool:
        assert self.profile == other.profile, "Cannot compare paths from different profiles"

        return self.internal_path < other.internal_path

    def __str__(self) -> str:
        return self.url

    def __repr__(self) -> str:
        return f"EPath({str(self)!r})"

    def __hash__(self) -> int:
        return hash((self.internal_path, self.profile))

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, EPath)
            and self.internal_path == other.internal_path
            and self.profile == other.profile
        )



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import functools
import os
import weakref
from typing import Callable, Protocol, Type, TypeVar

_after_in_child_fork_hooks = weakref.WeakKeyDictionary()
_after_in_parent_fork_hooks = weakref.WeakKeyDictionary()
_before_fork_hooks = weakref.WeakKeyDictionary()


T = TypeVar("T", bound=Callable[[], None])


def before_fork_hook(callable: Callable[[], None]):
    """
    Run function before the fork of a worker process. The function must be persistent.
    """
    # Make sure, that callable is a method of object
    assert getattr(callable, "__self__", None) is None, (
        f"Callable must not be a method: {callable.__name__}"
    )
    # print(f"Adding before_fork_hook for {callable.__name__}\n", end="")
    _before_fork_hooks[callable] = callable


def after_in_parent_fork_hook(callable: T):
    """
    Run function after the fork of a worker process. The function must be persistent.
    """
    # print(f"Adding after_in_child_fork_hook for {callable.__name__}\n", end="")
    assert getattr(callable, "__self__", None) is None, (
        f"Callable must not be a method: {callable.__name__}"
    )
    _after_in_parent_fork_hooks[callable] = callable


def after_in_child_fork_hook(callable: T):
    """
    Run function after the fork of a worker process. The function must be persistent.
    """
    # print(f"Adding after_in_child_fork_hook for {callable.__name__}\n", end="")
    assert getattr(callable, "__self__", None) is None, (
        f"Callable must not be a method: {callable.__name__}"
    )
    _after_in_child_fork_hooks[callable] = callable


class ForkMixin:
    """
    A mixin that runs a method after the fork of a worker process.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if getattr(self.__before_fork__, "__func__", None) is not ForkMixin.__before_fork__:
            _before_fork_hooks[self] = "__before_fork__"
        if (
            getattr(self.__after_in_child_fork__, "__func__", None)
            is not ForkMixin.__after_in_child_fork__
        ):
            _after_in_child_fork_hooks[self] = "__after_in_child_fork__"
        if (
            getattr(self.__after_in_parent_fork__, "__func__", None)
            is not ForkMixin.__after_in_parent_fork__
        ):
            _after_in_parent_fork_hooks[self] = "__after_in_parent_fork__"

    def __after_in_child_fork__(self):
        """
        A method that runs after the fork in the child process.
        """
        pass

    def __after_in_parent_fork__(self):
        """
        A method that runs after the fork in the parent process.
        """
        pass

    def __before_fork__(self):
        """
        A method that runs before the fork of a worker process.
        """
        pass


class ForkHookProtocol(Protocol):
    """
    A protocol that defines a method that runs before and after the fork of a worker process.
    """

    def __after_in_child_fork__(self):
        """
        A method that runs after the fork in the child process.
        """
        ...

    def __after_in_parent_fork__(self):
        """
        A method that runs after the fork in the parent process.
        """
        ...

    def __before_fork__(self):
        """
        A method that runs before the fork of a worker process.
        """
        ...


T_CLS = TypeVar("T_CLS", bound=Type[ForkHookProtocol])


def fork_hook_class(cls: T_CLS) -> T_CLS:
    """
    A decorator that runs a function after the fork of a worker process.
    """
    if hasattr(cls, "__init__"):
        orig_init = cls.__init__

        @functools.wraps(orig_init)
        def __init__(self, *args, **kwargs):
            _after_in_child_fork_hooks[self] = "__after_in_child_fork__"
            _after_in_parent_fork_hooks[self] = "__after_in_parent_fork__"
            _before_fork_hooks[self] = "__before_fork__"
            orig_init(self, *args, **kwargs)

        cls.__init__ = __init__
    else:

        def __init__(self, *args, **kwargs):
            _after_in_child_fork_hooks[cls] = "__after_in_child_fork__"
            _after_in_parent_fork_hooks[cls] = "__after_in_parent_fork__"
            _before_fork_hooks[cls] = "__before_fork__"
            cls(*args, **kwargs)

        cls.__init__ = __init__
    return cls


def _run_before_fork_hooks():
    """
    Run all the functions that were registered with the before_fork_hook decorator.
    """
    # print(f"Running before_fork_hooks for pid={os.getpid()}")
    for obj, hook in _before_fork_hooks.items():
        # print(f"Running before_fork_hook for {hook}\n", end="")
        if callable(hook):
            hook()
        else:
            getattr(obj, hook)()


def _run_after_in_child_fork_hooks():
    """
    Run all the functions that were registered with the after_in_child_fork_hook decorator.
    """
    # print(f"Running after_in_child_fork_hooks for pid={os.getpid()}")
    for obj, hook in _after_in_child_fork_hooks.items():
        # print(f"Running after_in_child_fork_hook for {hook}\n", end="")
        if callable(hook):
            hook()
        else:
            getattr(obj, hook)()


def _run_after_in_parent_fork_hooks():
    """
    Run all the functions that were registered with the after_in_parent_fork_hook decorator.
    """
    # print(f"Running after_in_parent_fork_hooks for pid={os.getpid()}")
    for obj, hook in _after_in_parent_fork_hooks.items():
        # print(f"Running after_in_parent_fork_hook for {hook}\n", end="")
        if callable(hook):
            hook()
        else:
            getattr(obj, hook)()


os.register_at_fork(
    before=_run_before_fork_hooks,
    after_in_child=_run_after_in_child_fork_hooks,
    after_in_parent=_run_after_in_parent_fork_hooks,
)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import List, Union

import torch

from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib.epath import EPath
from megatron.energon.errors import warn_deprecated
from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.webdataset import DefaultDecoderWebdatasetFactory


@edataclass
class InterleavedSample(Sample):
    """Sample type for interleaved media such as text with images."""

    #: The interleaved media (either torch.tensor for an image, or str for text)
    sequence: List[Union[torch.Tensor, str]]


class InterleavedWebdataset(DefaultDecoderWebdatasetFactory[InterleavedSample]):
    __sample_type__ = InterleavedSample

    def __init__(self, path: EPath, **kwargs):
        warn_deprecated(
            f"{type(self)} is deprecated, use the default instead and set the sample_type:\n"
            f"To convert, update your {path}/.nv-meta/dataset.yaml to:\n"
            f"# remove top-level __module__ and __class__\n"
            f"sample_type:\n"
            f"  __module__: megatron.energon\n"
            f"  __class__: {self.__sample_type__.__name__}\n"
            f"# Keep the remaining content"
        )
        super().__init__(path, **kwargs)
